"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    tests.unit.log_test
    ~~~~~~~~~~~~~~~~~~~

    Test salt's "hacked" logging
"""


import io
import logging

from salt._logging.handlers import StreamHandler
from salt._logging.impl import SaltLoggingClass
from tests.support.case import TestCase
from tests.support.helpers import TstSuiteLoggingHandler


class TestLog(TestCase):
    """
    Test several logging settings
    """

    def test_issue_2853_regex_TypeError(self):
        # Now, python's logging logger class is ours.
        # Let's make sure we have at least one instance
        log = SaltLoggingClass(__name__)

        # Test for a format which includes digits in name formatting.
        log_format = "[%(name)-15s] %(message)s"
        handler = TstSuiteLoggingHandler(format=log_format)
        log.addHandler(handler)

        # Trigger TstSuiteLoggingHandler.__enter__
        with handler:
            # Let's create another log instance to trigger salt's logging class
            # calculations.
            try:
                SaltLoggingClass("{}.with_digits".format(__name__))
            except Exception as err:  # pylint: disable=broad-except
                raise AssertionError(
                    "No exception should have been raised: {}".format(err)
                )

        # Remove the testing handler
        log.removeHandler(handler)

        # Test for a format which does not include digits in name formatting.
        log_format = "[%(name)s] %(message)s"
        handler = TstSuiteLoggingHandler(format=log_format)
        log.addHandler(handler)

        # Trigger TstSuiteLoggingHandler.__enter__
        with handler:
            # Let's create another log instance to trigger salt's logging class
            # calculations.
            try:
                SaltLoggingClass("{}.without_digits".format(__name__))
            except Exception as err:  # pylint: disable=broad-except
                raise AssertionError(
                    "No exception should have been raised: {}".format(err)
                )

            # Remove the testing handler
            log.removeHandler(handler)

    def test_exc_info_on_loglevel(self):
        def raise_exception_on_purpose():
            1 / 0  # pylint: disable=pointless-statement

        log = SaltLoggingClass(__name__)

        # Only stream2 should contain the traceback
        stream1 = io.StringIO()
        stream2 = io.StringIO()
        handler1 = StreamHandler(stream1)
        handler2 = StreamHandler(stream2)

        handler1.setLevel(logging.INFO)
        handler2.setLevel(logging.DEBUG)

        log.addHandler(handler1)
        log.addHandler(handler2)

        try:
            raise_exception_on_purpose()
        except ZeroDivisionError as exc:
            log.error(
                "Exception raised on purpose caught: ZeroDivisionError",
                exc_info_on_loglevel=logging.DEBUG,
            )

        try:
            self.assertIn(
                "Exception raised on purpose caught: ZeroDivisionError",
                stream1.getvalue(),
            )
            self.assertNotIn("Traceback (most recent call last)", stream1.getvalue())
            self.assertIn(
                "Exception raised on purpose caught: ZeroDivisionError",
                stream2.getvalue(),
            )
            self.assertIn("Traceback (most recent call last)", stream2.getvalue())
        finally:
            log.removeHandler(handler1)
            log.removeHandler(handler2)

        # Both streams should contain the traceback
        stream1 = io.StringIO()
        stream2 = io.StringIO()
        handler1 = StreamHandler(stream1)
        handler2 = StreamHandler(stream2)

        handler1.setLevel(logging.INFO)
        handler2.setLevel(logging.DEBUG)

        log.addHandler(handler1)
        log.addHandler(handler2)

        try:
            raise_exception_on_purpose()
        except ZeroDivisionError as exc:
            log.error(
                "Exception raised on purpose caught: ZeroDivisionError",
                exc_info_on_loglevel=logging.INFO,
            )

        try:
            self.assertIn(
                "Exception raised on purpose caught: ZeroDivisionError",
                stream1.getvalue(),
            )
            self.assertIn("Traceback (most recent call last)", stream1.getvalue())
            self.assertIn(
                "Exception raised on purpose caught: ZeroDivisionError",
                stream2.getvalue(),
            )
            self.assertIn("Traceback (most recent call last)", stream2.getvalue())
        finally:
            log.removeHandler(handler1)
            log.removeHandler(handler2)

        # No streams should contain the traceback
        stream1 = io.StringIO()
        stream2 = io.StringIO()
        handler1 = StreamHandler(stream1)
        handler2 = StreamHandler(stream2)

        handler1.setLevel(logging.ERROR)
        handler2.setLevel(logging.INFO)

        log.addHandler(handler1)
        log.addHandler(handler2)

        try:
            raise_exception_on_purpose()
        except ZeroDivisionError as exc:
            log.error(
                "Exception raised on purpose caught: ZeroDivisionError",
                exc_info_on_loglevel=logging.DEBUG,
            )

        try:
            self.assertIn(
                "Exception raised on purpose caught: ZeroDivisionError",
                stream1.getvalue(),
            )
            self.assertNotIn("Traceback (most recent call last)", stream1.getvalue())
            self.assertIn(
                "Exception raised on purpose caught: ZeroDivisionError",
                stream2.getvalue(),
            )
            self.assertNotIn("Traceback (most recent call last)", stream2.getvalue())
        finally:
            log.removeHandler(handler1)
            log.removeHandler(handler2)



import pytest
import salt.states.win_powercfg as powercfg
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {powercfg: {}}


def test_set_monitor():
    """
    Test to make sure we can set the monitor timeout value
    """
    ret = {
        "changes": {"monitor": {"ac": {"new": 0, "old": 45}}},
        "comment": "Monitor timeout on AC power set to 0",
        "name": "monitor",
        "result": True,
    }
    get_monitor_side_effect = MagicMock(
        side_effect=[{"ac": 45, "dc": 22}, {"ac": 0, "dc": 22}]
    )
    with patch.dict(
        powercfg.__salt__,
        {
            "powercfg.get_monitor_timeout": get_monitor_side_effect,
            "powercfg.set_monitor_timeout": MagicMock(return_value=True),
        },
    ):
        with patch.dict(powercfg.__opts__, {"test": False}):
            assert powercfg.set_timeout("monitor", 0) == ret


def test_set_monitor_already_set():
    """
    Test to make sure we can set the monitor timeout value
    """
    ret = {
        "changes": {},
        "comment": "Monitor timeout on AC power is already set to 0",
        "name": "monitor",
        "result": True,
    }
    monitor_val = MagicMock(return_value={"ac": 0, "dc": 0})
    with patch.dict(powercfg.__salt__, {"powercfg.get_monitor_timeout": monitor_val}):
        assert powercfg.set_timeout("monitor", 0) == ret


def test_set_monitor_test_true_with_change():
    """
    Test to make sure set monitor works correctly with test=True with
    changes
    """
    ret = {
        "changes": {},
        "comment": "Monitor timeout on AC power will be set to 0",
        "name": "monitor",
        "result": None,
    }
    get_monitor_return_value = MagicMock(return_value={"ac": 45, "dc": 22})
    with patch.dict(
        powercfg.__salt__,
        {"powercfg.get_monitor_timeout": get_monitor_return_value},
    ):
        with patch.dict(powercfg.__opts__, {"test": True}):
            assert powercfg.set_timeout("monitor", 0) == ret


def test_fail_invalid_setting():
    """
    Test to make sure we can set the monitor timeout value
    """
    ret = {
        "changes": {},
        "comment": '"fakesetting" is not a valid setting',
        "name": "fakesetting",
        "result": False,
    }
    assert powercfg.set_timeout("fakesetting", 0) == ret


def test_fail_invalid_power():
    """
    Test to make sure we can set the monitor timeout value
    """
    ret = {
        "changes": {},
        "comment": '"fakepower" is not a power type',
        "name": "monitor",
        "result": False,
    }
    assert powercfg.set_timeout("monitor", 0, power="fakepower") == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.rdp as rdp
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class RdpTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.rdp
    """

    def setup_loader_modules(self):
        return {rdp: {}}

    # 'enable' function tests: 1

    def test_enable(self):
        """
        Test if it enables RDP the service on the server
        """
        mock = MagicMock(return_value=True)
        with patch.dict(rdp.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.rdp._parse_return_code_powershell", MagicMock(return_value=0)
        ):
            self.assertTrue(rdp.enable())

    # 'disable' function tests: 1

    def test_disable(self):
        """
        Test if it disables RDP the service on the server
        """
        mock = MagicMock(return_value=True)
        with patch.dict(rdp.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.rdp._parse_return_code_powershell", MagicMock(return_value=0)
        ):
            self.assertTrue(rdp.disable())

    # 'status' function tests: 1

    def test_status(self):
        """
        Test if it shows rdp is enabled on the server
        """
        mock = MagicMock(return_value="1")
        with patch.dict(rdp.__salt__, {"cmd.run": mock}):
            self.assertTrue(rdp.status())



"""
Management of APT/DNF/YUM/Zypper package repos
==============================================

States for managing software package repositories on Linux distros. Supported
package managers are APT, DNF, YUM and Zypper. Here is some example SLS:

.. code-block:: yaml

    base:
      pkgrepo.managed:
        - humanname: CentOS-$releasever - Base
        - mirrorlist: http://mirrorlist.centos.org/?release=$releasever&arch=$basearch&repo=os
        - comments:
            - 'http://mirror.centos.org/centos/$releasever/os/$basearch/'
        - gpgcheck: 1
        - gpgkey: file:///etc/pki/rpm-gpg/RPM-GPG-KEY-CentOS-6

.. code-block:: yaml

    base:
      pkgrepo.managed:
        - humanname: Logstash PPA
        - name: deb http://ppa.launchpad.net/wolfnet/logstash/ubuntu precise main
        - dist: precise
        - file: /etc/apt/sources.list.d/logstash.list
        - keyid: 28B04E4A
        - keyserver: keyserver.ubuntu.com
        - require_in:
          - pkg: logstash

      pkg.latest:
        - name: logstash
        - refresh: True

.. code-block:: yaml

    base:
      pkgrepo.managed:
        - humanname: deb-multimedia
        - name: deb http://www.deb-multimedia.org stable main
        - file: /etc/apt/sources.list.d/deb-multimedia.list
        - key_url: salt://deb-multimedia/files/marillat.pub

.. code-block:: yaml

    base:
      pkgrepo.managed:
        - humanname: Google Chrome
        - name: deb http://dl.google.com/linux/chrome/deb/ stable main
        - dist: stable
        - file: /etc/apt/sources.list.d/chrome-browser.list
        - require_in:
          - pkg: google-chrome-stable
        - gpgcheck: 1
        - key_url: https://dl-ssl.google.com/linux/linux_signing_key.pub

.. code-block:: yaml

    base:
      pkgrepo.managed:
        - ppa: wolfnet/logstash
      pkg.latest:
        - name: logstash
        - refresh: True

.. note::

    On Ubuntu systems, the ``python-software-properties`` package should be
    installed for better support of PPA repositories. To check if this package
    is installed, run ``dpkg -l python-software-properties``.

    On Ubuntu & Debian systems, the ``python-apt`` package is required to be
    installed. To check if this package is installed, run ``dpkg -l python-apt``.
    ``python-apt`` will need to be manually installed if it is not present.

.. code-block:: yaml

    hello-copr:
        pkgrepo.managed:
            - copr: mymindstorm/hello
        pkg.installed:
            - name: hello

"""


import sys

import salt.utils.data
import salt.utils.files
import salt.utils.pkg.deb
import salt.utils.pkg.rpm
import salt.utils.versions
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.state import STATE_INTERNAL_KEYWORDS as _STATE_INTERNAL_KEYWORDS


def __virtual__():
    """
    Only load if modifying repos is available for this package type
    """
    return "pkg.mod_repo" in __salt__


def managed(name, ppa=None, copr=None, **kwargs):
    """
    This state manages software package repositories. Currently, :mod:`yum
    <salt.modules.yumpkg>`, :mod:`apt <salt.modules.aptpkg>`, and :mod:`zypper
    <salt.modules.zypper>` repositories are supported.

    **YUM/DNF/ZYPPER-BASED SYSTEMS**

    .. note::
        One of ``baseurl`` or ``mirrorlist`` below is required. Additionally,
        note that this state is not presently capable of managing more than one
        repo in a single repo file, so each instance of this state will manage
        a single repo file containing the configuration for a single repo.

    name
        This value will be used in two ways: Firstly, it will be the repo ID,
        as seen in the entry in square brackets (e.g. ``[foo]``) for a given
        repo. Secondly, it will be the name of the file as stored in
        /etc/yum.repos.d (e.g. ``/etc/yum.repos.d/foo.conf``).

    enabled : True
        Whether the repo is enabled or not. Can be specified as ``True``/``False`` or
        ``1``/``0``.

    disabled : False
        Included to reduce confusion due to APT's use of the ``disabled``
        argument. If this is passed for a YUM/DNF/Zypper-based distro, then the
        reverse will be passed as ``enabled``. For example passing
        ``disabled=True`` will assume ``enabled=False``.

    copr
        Fedora and RedHat based distributions only. Use community packages
        outside of the main package repository.

        .. versionadded:: 3002

    humanname
        This is used as the ``name`` value in the repo file in
        ``/etc/yum.repos.d/`` (or ``/etc/zypp/repos.d`` for SUSE distros).

    baseurl
        The URL to a yum repository

    mirrorlist
        A URL which points to a file containing a collection of baseurls

    comments
        Sometimes you want to supply additional information, but not as
        enabled configuration. Anything supplied for this list will be saved
        in the repo configuration with a comment marker (#) in front.

    gpgautoimport
        Only valid for Zypper package manager. If set to ``True``, automatically
        trust and import the new repository signing key. The key should be
        specified with ``gpgkey`` parameter. See details below.

    Additional configuration values seen in YUM/DNF/Zypper repo files, such as
    ``gpgkey`` or ``gpgcheck``, will be used directly as key-value pairs.
    For example:

    .. code-block:: yaml

        foo:
          pkgrepo.managed:
            - humanname: Personal repo for foo
            - baseurl: https://mydomain.tld/repo/foo/$releasever/$basearch
            - gpgkey: file:///etc/pki/rpm-gpg/foo-signing-key
            - gpgcheck: 1


    **APT-BASED SYSTEMS**

    ppa
        On Ubuntu, you can take advantage of Personal Package Archives on
        Launchpad simply by specifying the user and archive name. The keyid
        will be queried from launchpad and everything else is set
        automatically. You can override any of the below settings by simply
        setting them as you would normally. For example:

        .. code-block:: yaml

            logstash-ppa:
              pkgrepo.managed:
                - ppa: wolfnet/logstash

    ppa_auth
        For Ubuntu PPAs there can be private PPAs that require authentication
        to access. For these PPAs the username/password can be passed as an
        HTTP Basic style username/password combination.

        .. code-block:: yaml

            logstash-ppa:
              pkgrepo.managed:
                - ppa: wolfnet/logstash
                - ppa_auth: username:password

    name
        On apt-based systems this must be the complete entry as it would be
        seen in the ``sources.list`` file. This can have a limited subset of
        components (e.g. ``main``) which can be added/modified with the
        ``comps`` option.

        .. code-block:: yaml

            precise-repo:
              pkgrepo.managed:
                - name: deb http://us.archive.ubuntu.com/ubuntu precise main

        .. note::

            The above example is intended as a more readable way of configuring
            the SLS, it is equivalent to the following:

            .. code-block:: yaml

                'deb http://us.archive.ubuntu.com/ubuntu precise main':
                  pkgrepo.managed

    disabled : False
        Toggles whether or not the repo is used for resolving dependencies
        and/or installing packages.

    enabled : True
        Included to reduce confusion due to YUM/DNF/Zypper's use of the
        ``enabled`` argument. If this is passed for an APT-based distro, then
        the reverse will be passed as ``disabled``. For example, passing
        ``enabled=False`` will assume ``disabled=False``.

    architectures
        On apt-based systems, ``architectures`` can restrict the available
        architectures that the repository provides (e.g. only ``amd64``).
        ``architectures`` should be a comma-separated list.

    comps
        On apt-based systems, comps dictate the types of packages to be
        installed from the repository (e.g. ``main``, ``nonfree``, ...).  For
        purposes of this, ``comps`` should be a comma-separated list.

    file
       The filename for the ``*.list`` that the repository is configured in.
       It is important to include the full-path AND make sure it is in
       a directory that APT will look in when handling packages

    dist
       This dictates the release of the distro the packages should be built
       for.  (e.g. ``unstable``). This option is rarely needed.

    keyid
       The KeyID or a list of KeyIDs of the GPG key to install.
       This option also requires the ``keyserver`` option to be set.

    keyserver
       This is the name of the keyserver to retrieve GPG keys from. The
       ``keyid`` option must also be set for this option to work.

    key_url
       URL to retrieve a GPG key from. Allows the usage of ``http://``,
       ``https://`` as well as ``salt://``.

       .. note::

           Use either ``keyid``/``keyserver`` or ``key_url``, but not both.

    key_text
       The string representation of the GPG key to install.

       .. versionadded:: 2018.3.0

       .. note::

           Use either ``keyid``/``keyserver``, ``key_url``, or ``key_text`` but
           not more than one method.

    consolidate : False
       If set to ``True``, this will consolidate all sources definitions to the
       ``sources.list`` file, cleanup the now unused files, consolidate components
       (e.g. ``main``) for the same URI, type, and architecture to a single line,
       and finally remove comments from the ``sources.list`` file.  The consolidation
       will run every time the state is processed. The option only needs to be
       set on one repo managed by Salt to take effect.

    clean_file : False
       If set to ``True``, empty the file before configuring the defined repository

       .. note::
           Use with care. This can be dangerous if multiple sources are
           configured in the same file.

       .. versionadded:: 2015.8.0

    refresh : True
       If set to ``False`` this will skip refreshing the apt package database
       on Debian based systems.

    refresh_db : True
       .. deprecated:: 2018.3.0
           Use ``refresh`` instead.

    require_in
       Set this to a list of :mod:`pkg.installed <salt.states.pkg.installed>` or
       :mod:`pkg.latest <salt.states.pkg.latest>` to trigger the
       running of ``apt-get update`` prior to attempting to install these
       packages. Setting a require in the pkg state will not work for this.
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if "pkg.get_repo" not in __salt__:
        ret["result"] = False
        ret["comment"] = "Repo management not implemented on this platform"
        return ret

    if "key_url" in kwargs and ("keyid" in kwargs or "keyserver" in kwargs):
        ret["result"] = False
        ret[
            "comment"
        ] = 'You may not use both "keyid"/"keyserver" and "key_url" argument.'

    if "key_text" in kwargs and ("keyid" in kwargs or "keyserver" in kwargs):
        ret["result"] = False
        ret[
            "comment"
        ] = 'You may not use both "keyid"/"keyserver" and "key_text" argument.'
    if "key_text" in kwargs and ("key_url" in kwargs):
        ret["result"] = False
        ret["comment"] = 'You may not use both "key_url" and "key_text" argument.'

    if "repo" in kwargs:
        ret["result"] = False
        ret["comment"] = (
            "'repo' is not a supported argument for this "
            "state. The 'name' argument is probably what was "
            "intended."
        )
        return ret

    enabled = kwargs.pop("enabled", None)
    disabled = kwargs.pop("disabled", None)

    if enabled is not None and disabled is not None:
        ret["result"] = False
        ret["comment"] = "Only one of enabled/disabled is allowed"
        return ret
    elif enabled is None and disabled is None:
        # If neither argument was passed we assume the repo will be enabled
        enabled = True

    repo = name
    if __grains__["os"] in ("Ubuntu", "Mint"):
        if ppa is not None:
            # overload the name/repo value for PPAs cleanly
            # this allows us to have one code-path for PPAs
            try:
                repo = ":".join(("ppa", ppa))
            except TypeError:
                repo = ":".join(("ppa", str(ppa)))

        kwargs["disabled"] = (
            not salt.utils.data.is_true(enabled)
            if enabled is not None
            else salt.utils.data.is_true(disabled)
        )

    elif __grains__["os_family"] in ("RedHat", "Suse"):
        if __grains__["os_family"] in "RedHat":
            if copr is not None:
                repo = ":".join(("copr", copr))
                kwargs["name"] = name

        if "humanname" in kwargs:
            kwargs["name"] = kwargs.pop("humanname")
        if "name" not in kwargs:
            # Fall back to the repo name if humanname not provided
            kwargs["name"] = repo

        kwargs["enabled"] = (
            not salt.utils.data.is_true(disabled)
            if disabled is not None
            else salt.utils.data.is_true(enabled)
        )

    elif __grains__["os_family"] in ("NILinuxRT", "Poky"):
        # opkg is the pkg virtual
        kwargs["enabled"] = (
            not salt.utils.data.is_true(disabled)
            if disabled is not None
            else salt.utils.data.is_true(enabled)
        )

    for kwarg in _STATE_INTERNAL_KEYWORDS:
        kwargs.pop(kwarg, None)

    try:
        pre = __salt__["pkg.get_repo"](repo=repo, **kwargs)
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = "Failed to examine repo '{}': {}".format(name, exc)
        return ret

    # This is because of how apt-sources works. This pushes distro logic
    # out of the state itself and into a module that it makes more sense
    # to use. Most package providers will simply return the data provided
    # it doesn't require any "specialized" data massaging.
    if "pkg.expand_repo_def" in __salt__:
        sanitizedkwargs = __salt__["pkg.expand_repo_def"](repo=repo, **kwargs)
    else:
        sanitizedkwargs = kwargs

    if __grains__["os_family"] == "Debian":
        repo = salt.utils.pkg.deb.strip_uri(repo)

    if pre:
        # 22412: Remove file attribute in case same repo is set up multiple times but with different files
        pre.pop("file", None)
        sanitizedkwargs.pop("file", None)
        for kwarg in sanitizedkwargs:
            if kwarg not in pre:
                if kwarg == "enabled":
                    # On a RedHat-based OS, 'enabled' is assumed to be true if
                    # not explicitly set, so we don't need to update the repo
                    # if it's desired to be enabled and the 'enabled' key is
                    # missing from the repo definition
                    if __grains__["os_family"] == "RedHat":
                        if not salt.utils.data.is_true(sanitizedkwargs[kwarg]):
                            break
                    else:
                        break
                else:
                    break
            elif kwarg in ("comps", "key_url"):
                if sorted(sanitizedkwargs[kwarg]) != sorted(pre[kwarg]):
                    break
            elif kwarg == "line" and __grains__["os_family"] == "Debian":
                # split the line and sort everything after the URL
                sanitizedsplit = sanitizedkwargs[kwarg].split()
                sanitizedsplit[3:] = sorted(sanitizedsplit[3:])
                reposplit, _, pre_comments = [
                    x.strip() for x in pre[kwarg].partition("#")
                ]
                reposplit = reposplit.split()
                reposplit[3:] = sorted(reposplit[3:])
                if sanitizedsplit != reposplit:
                    break
                if "comments" in kwargs:
                    post_comments = salt.utils.pkg.deb.combine_comments(
                        kwargs["comments"]
                    )
                    if pre_comments != post_comments:
                        break
            elif kwarg == "comments" and __grains__["os_family"] == "RedHat":
                precomments = salt.utils.pkg.rpm.combine_comments(pre[kwarg])
                kwargcomments = salt.utils.pkg.rpm.combine_comments(
                    sanitizedkwargs[kwarg]
                )
                if precomments != kwargcomments:
                    break
            elif kwarg == "architectures" and sanitizedkwargs[kwarg]:
                if set(sanitizedkwargs[kwarg]) != set(pre[kwarg]):
                    break
            else:
                if __grains__["os_family"] in ("RedHat", "Suse") and any(
                    isinstance(x, bool) for x in (sanitizedkwargs[kwarg], pre[kwarg])
                ):
                    # This check disambiguates 1/0 from True/False
                    if salt.utils.data.is_true(
                        sanitizedkwargs[kwarg]
                    ) != salt.utils.data.is_true(pre[kwarg]):
                        break
                else:
                    if str(sanitizedkwargs[kwarg]) != str(pre[kwarg]):
                        break
        else:
            ret["result"] = True
            ret["comment"] = "Package repo '{}' already configured".format(name)
            return ret

    if __opts__["test"]:
        ret["comment"] = (
            "Package repo '{}' would be configured. This may cause pkg "
            "states to behave differently than stated if this action is "
            "repeated without test=True, due to the differences in the "
            "configured repositories.".format(name)
        )
        if pre:
            for kwarg in sanitizedkwargs:
                if sanitizedkwargs.get(kwarg) != pre.get(kwarg):
                    ret["changes"][kwarg] = {
                        "new": sanitizedkwargs.get(kwarg),
                        "old": pre.get(kwarg),
                    }
        else:
            ret["changes"]["repo"] = name
        return ret

    # empty file before configure
    if kwargs.get("clean_file", False):
        with salt.utils.files.fopen(kwargs["file"], "w"):
            pass

    try:
        if __grains__["os_family"] == "Debian":
            __salt__["pkg.mod_repo"](repo, saltenv=__env__, **kwargs)
        else:
            __salt__["pkg.mod_repo"](repo, **kwargs)
    except Exception as exc:  # pylint: disable=broad-except
        # This is another way to pass information back from the mod_repo
        # function.
        ret["result"] = False
        ret["comment"] = "Failed to configure repo '{}': {}".format(name, exc)
        return ret

    try:
        post = __salt__["pkg.get_repo"](repo=repo, **kwargs)
        if pre:
            for kwarg in sanitizedkwargs:
                if post.get(kwarg) != pre.get(kwarg):
                    ret["changes"][kwarg] = {
                        "new": post.get(kwarg),
                        "old": pre.get(kwarg),
                    }
        else:
            ret["changes"] = {"repo": repo}

        ret["result"] = True
        ret["comment"] = "Configured package repo '{}'".format(name)
    except Exception as exc:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = "Failed to confirm config of repo '{}': {}".format(name, exc)

    # Clear cache of available packages, if present, since changes to the
    # repositories may change the packages that are available.
    if ret["changes"]:
        sys.modules[__salt__["test.ping"].__module__].__context__.pop(
            "pkg._avail", None
        )

    return ret


def absent(name, **kwargs):
    """
    This function deletes the specified repo on the system, if it exists. It
    is essentially a wrapper around :mod:`pkg.del_repo <salt.modules.pkg.del_repo>`.

    name
        The name of the package repo, as it would be referred to when running
        the regular package manager commands.

    **FEDORA/REDHAT-SPECIFIC OPTIONS**

    copr
        Use community packages outside of the main package repository.

        .. versionadded:: 3002

        .. code-block:: yaml

            hello-copr:
                pkgrepo.absent:
                  - copr: mymindstorm/hello

    **UBUNTU-SPECIFIC OPTIONS**

    ppa
        On Ubuntu, you can take advantage of Personal Package Archives on
        Launchpad simply by specifying the user and archive name.

        .. code-block:: yaml

            logstash-ppa:
              pkgrepo.absent:
                - ppa: wolfnet/logstash

    ppa_auth
        For Ubuntu PPAs there can be private PPAs that require authentication
        to access. For these PPAs the username/password can be specified.  This
        is required for matching if the name format uses the ``ppa:`` specifier
        and is private (requires username/password to access, which is encoded
        in the URI).

        .. code-block:: yaml

            logstash-ppa:
              pkgrepo.absent:
                - ppa: wolfnet/logstash
                - ppa_auth: username:password

    keyid
        If passed, then the GPG key corresponding to the passed KeyID will also
        be removed.

    keyid_ppa : False
        If set to ``True``, the GPG key's ID will be looked up from
        ppa.launchpad.net and removed, and the ``keyid`` argument will be
        ignored.

        .. note::
            This option will be disregarded unless the ``ppa`` argument is
            present.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if "ppa" in kwargs and __grains__["os"] in ("Ubuntu", "Mint"):
        name = kwargs.pop("ppa")
        if not name.startswith("ppa:"):
            name = "ppa:" + name

    if "copr" in kwargs and __grains__["os_family"] in "RedHat":
        name = kwargs.pop("copr")
        if not name.startswith("copr:"):
            name = "copr:" + name

    remove_key = any(kwargs.get(x) is not None for x in ("keyid", "keyid_ppa"))
    if remove_key and "pkg.del_repo_key" not in __salt__:
        ret["result"] = False
        ret["comment"] = "Repo key management is not implemented for this platform"
        return ret

    try:
        repo = __salt__["pkg.get_repo"](name, **kwargs)
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = "Failed to configure repo '{}': {}".format(name, exc)
        return ret

    if not repo:
        ret["comment"] = "Package repo {} is absent".format(name)
        ret["result"] = True
        return ret

    if __opts__["test"]:
        ret["comment"] = (
            "Package repo '{}' will be removed. This may "
            "cause pkg states to behave differently than stated "
            "if this action is repeated without test=True, due "
            "to the differences in the configured repositories.".format(name)
        )
        return ret

    try:
        __salt__["pkg.del_repo"](repo=name, **kwargs)
    except (CommandExecutionError, SaltInvocationError) as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret

    repos = __salt__["pkg.list_repos"]()
    if name not in repos:
        ret["changes"]["repo"] = name
        ret["comment"] = "Removed repo {}".format(name)

        if not remove_key:
            ret["result"] = True
        else:
            try:
                removed_keyid = __salt__["pkg.del_repo_key"](name, **kwargs)
            except (CommandExecutionError, SaltInvocationError) as exc:
                ret["result"] = False
                ret["comment"] += ", but failed to remove key: {}".format(exc)
            else:
                ret["result"] = True
                ret["changes"]["keyid"] = removed_keyid
                ret["comment"] += ", and keyid {}".format(removed_keyid)
    else:
        ret["result"] = False
        ret["comment"] = "Failed to remove repo {}".format(name)

    return ret



"""
Monit service module. This module will create a monit type
service watcher.
"""

import re

import salt.utils.path

# Function alias to make sure not to shadow built-in's
__func_alias__ = {
    "id_": "id",
    "reload_": "reload",
}


def __virtual__():
    if salt.utils.path.which("monit") is not None:
        # The monit binary exists, let the module load
        return True
    return (
        False,
        "The monit execution module cannot be loaded: the monit binary is not in the"
        " path.",
    )


def start(name):
    """

    CLI Example:

    .. code-block:: bash

        salt '*' monit.start <service name>
    """
    cmd = "monit start {}".format(name)

    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def stop(name):
    """
    Stops service via monit

    CLI Example:

    .. code-block:: bash

        salt '*' monit.stop <service name>
    """
    cmd = "monit stop {}".format(name)

    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def restart(name):
    """
    Restart service via monit

    CLI Example:

    .. code-block:: bash

        salt '*' monit.restart <service name>
    """
    cmd = "monit restart {}".format(name)

    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def unmonitor(name):
    """
    Unmonitor service via monit

    CLI Example:

    .. code-block:: bash

        salt '*' monit.unmonitor <service name>
    """
    cmd = "monit unmonitor {}".format(name)

    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def monitor(name):
    """
    monitor service via monit

    CLI Example:

    .. code-block:: bash

        salt '*' monit.monitor <service name>
    """
    cmd = "monit monitor {}".format(name)

    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def summary(svc_name=""):
    """
    Display a summary from monit

    CLI Example:

    .. code-block:: bash

        salt '*' monit.summary
        salt '*' monit.summary <service name>
    """
    ret = {}
    cmd = "monit summary"
    res = __salt__["cmd.run"](cmd).splitlines()
    for line in res:
        if "daemon is not running" in line:
            return dict(monit="daemon is not running", result=False)
        elif not line or svc_name not in line or "The Monit daemon" in line:
            continue
        else:
            parts = line.split("'")
            if len(parts) == 3:
                resource, name, status_ = (parts[0].strip(), parts[1], parts[2].strip())
                if svc_name != "" and svc_name != name:
                    continue
                if resource not in ret:
                    ret[resource] = {}
                ret[resource][name] = status_
    return ret


def status(svc_name=""):
    """
    Display a process status from monit

    CLI Example:

    .. code-block:: bash

        salt '*' monit.status
        salt '*' monit.status <service name>
    """

    cmd = "monit status"
    res = __salt__["cmd.run"](cmd)

    # Monit uses a different separator since 5.18.0
    if version() < "5.18.0":
        fieldlength = 33
    else:
        fieldlength = 28

    separator = 3 + fieldlength
    prostr = "Process" + " " * fieldlength

    s = res.replace("Process", prostr).replace("'", "").split("\n\n")
    entries = {}
    for process in s[1:-1]:
        pro = process.splitlines()
        tmp = {}
        for items in pro:
            key = items[:separator].strip()
            tmp[key] = items[separator - 1 :].strip()
        entries[pro[0].split()[1]] = tmp
    if svc_name == "":
        ret = entries
    else:
        ret = entries.get(svc_name, "No such service")
    return ret


def reload_():
    """
    .. versionadded:: 2016.3.0

    Reload monit configuration

    CLI Example:

    .. code-block:: bash

        salt '*' monit.reload
    """
    cmd = "monit reload"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def configtest():
    """
    .. versionadded:: 2016.3.0

    Test monit configuration syntax

    CLI Example:

    .. code-block:: bash

        salt '*' monit.configtest
    """
    ret = {}
    cmd = "monit -t"
    out = __salt__["cmd.run_all"](cmd)

    if out["retcode"] != 0:
        ret["comment"] = "Syntax Error"
        ret["stderr"] = out["stderr"]
        ret["result"] = False
        return ret

    ret["comment"] = "Syntax OK"
    ret["stdout"] = out["stdout"]
    ret["result"] = True
    return ret


def version():
    """
    .. versionadded:: 2016.3.0

    Return version from monit -V

    CLI Example:

    .. code-block:: bash

        salt '*' monit.version
    """
    cmd = "monit -V"
    out = __salt__["cmd.run"](cmd).splitlines()
    ret = out[0].split()
    return ret[-1]


def id_(reset=False):
    """
    .. versionadded:: 2016.3.0

    Return monit unique id.

    reset : False
        Reset current id and generate a new id when it's True.

    CLI Example:

    .. code-block:: bash

        salt '*' monit.id [reset=True]
    """
    if reset:
        id_pattern = re.compile(r"Monit id (?P<id>[^ ]+)")
        cmd = "echo y|monit -r"
        out = __salt__["cmd.run_all"](cmd, python_shell=True)
        ret = id_pattern.search(out["stdout"]).group("id")
        return ret if ret else False
    else:
        cmd = "monit -i"
        out = __salt__["cmd.run"](cmd)
        ret = out.split(":")[-1].strip()
    return ret


def validate():
    """
    .. versionadded:: 2016.3.0

    Check all services

    CLI Example:

    .. code-block:: bash

        salt '*' monit.validate
    """
    cmd = "monit validate"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)



import salt.modules.vmctl as vmctl
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class VmctlTestCase(TestCase, LoaderModuleMockMixin):
    """
    test modules.vmctl functions
    """

    def setup_loader_modules(self):
        return {vmctl: {}}

    def test_create_disk(self):
        """
        Tests creating a new disk image.
        """
        ret = {}
        ret["stdout"] = "vmctl: imagefile created"
        ret["stderr"] = ""
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertTrue(vmctl.create_disk("/path/to/disk.img", "1G"))

    def test_load(self):
        """
        Tests loading a configuration file.
        """
        ret = {}
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertTrue(vmctl.load("/etc/vm.switches.conf"))

    def test_reload(self):
        """
        Tests reloading the configuration.
        """
        ret = {}
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertTrue(vmctl.reload())

    def test_reset(self):
        """
        Tests resetting VMM.
        """
        ret = {}
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            res = vmctl.reset()
            mock_cmd.assert_called_once_with(
                ["vmctl", "reset"], output_loglevel="trace", python_shell=False
            )
            self.assertTrue(res)

    def test_reset_vms(self):
        """
        Tests resetting VMs.
        """
        ret = {}
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            res = vmctl.reset(vms=True)
            mock_cmd.assert_called_once_with(
                ["vmctl", "reset", "vms"], output_loglevel="trace", python_shell=False
            )
            self.assertTrue(res)

    def test_reset_switches(self):
        """
        Tests resetting switches.
        """
        ret = {}
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            res = vmctl.reset(switches=True)
            mock_cmd.assert_called_once_with(
                ["vmctl", "reset", "switches"],
                output_loglevel="trace",
                python_shell=False,
            )
            self.assertTrue(res)

    def test_reset_all(self):
        """
        Tests resetting all.
        """
        ret = {}
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            res = vmctl.reset(all=True)
            mock_cmd.assert_called_once_with(
                ["vmctl", "reset", "all"], output_loglevel="trace", python_shell=False
            )
            self.assertTrue(res)

    def test_start_existing_vm(self):
        """
        Tests starting a VM that is already defined.
        """
        ret = {}
        ret["stderr"] = "vmctl: started vm 4 successfully, tty /dev/ttyp4"
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        expected = {"changes": True, "console": "/dev/ttyp4"}
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertDictEqual(expected, vmctl.start("4"))

    def test_start_new_vm(self):
        """
        Tests starting a new VM.
        """
        ret = {}
        ret["stderr"] = "vmctl: started vm 4 successfully, tty /dev/ttyp4"
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        mock_status = MagicMock(return_value={})
        expected = {"changes": True, "console": "/dev/ttyp4"}
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            with patch("salt.modules.vmctl.status", mock_status):
                res = vmctl.start("web1", bootpath="/bsd.rd", nics=2, disk="/disk.img")
                mock_cmd.assert_called_once_with(
                    [
                        "vmctl",
                        "start",
                        "web1",
                        "-i 2",
                        "-b",
                        "/bsd.rd",
                        "-d",
                        "/disk.img",
                    ],
                    output_loglevel="trace",
                    python_shell=False,
                )
                self.assertDictEqual(expected, res)

    def test_status(self):
        """
        Tests getting status for all VMs.
        """
        ret = {}
        ret["stdout"] = (
            "   ID   PID VCPUS  MAXMEM  CURMEM     TTY        OWNER NAME\n"
            "    1   123     1    2.9G     150M  ttyp5       john web1 - stopping\n"
            "    2   456     1    512M     301M  ttyp4       paul web2\n"
            "    3     -     1    512M       -       -       george web3\n"
        )
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        expected = {
            "web1": {
                "curmem": "150M",
                "id": "1",
                "maxmem": "2.9G",
                "owner": "john",
                "pid": "123",
                "state": "stopping",
                "tty": "ttyp5",
                "vcpus": "1",
            },
            "web2": {
                "curmem": "301M",
                "id": "2",
                "maxmem": "512M",
                "owner": "paul",
                "pid": "456",
                "state": "running",
                "tty": "ttyp4",
                "vcpus": "1",
            },
            "web3": {
                "curmem": "-",
                "id": "3",
                "maxmem": "512M",
                "owner": "george",
                "pid": "-",
                "state": "stopped",
                "tty": "-",
                "vcpus": "1",
            },
        }
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertEqual(expected, vmctl.status())

    def test_status_single(self):
        """
        Tests getting status for a single VM.
        """
        ret = {}
        ret["stdout"] = (
            "   ID   PID VCPUS  MAXMEM  CURMEM     TTY        OWNER NAME\n"
            "    1   123     1    2.9G     150M  ttyp5       ringo web4\n"
            "    2     -     1    512M       -       -       george web3\n"
        )
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        expected = {
            "web4": {
                "curmem": "150M",
                "id": "1",
                "maxmem": "2.9G",
                "owner": "ringo",
                "pid": "123",
                "state": "running",
                "tty": "ttyp5",
                "vcpus": "1",
            },
        }
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertEqual(expected, vmctl.status("web4"))

    def test_stop_when_running(self):
        """
        Tests stopping a VM that is running.
        """
        ret = {}
        ret["stdout"] = ""
        ret["stderr"] = "vmctl: sent request to terminate vm 14"
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            res = vmctl.stop("web1")
            mock_cmd.assert_called_once_with(
                ["vmctl", "stop", "web1"], output_loglevel="trace", python_shell=False
            )
            self.assertTrue(res["changes"])

    def test_stop_when_stopped(self):
        """
        Tests stopping a VM that is already stopped/stopping.
        """
        ret = {}
        ret["stdout"] = ""
        ret["stderr"] = "vmctl: terminate vm command failed: Invalid argument"
        ret["retcode"] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
            res = vmctl.stop("web1")
            mock_cmd.assert_called_once_with(
                ["vmctl", "stop", "web1"], output_loglevel="trace", python_shell=False
            )
            self.assertFalse(res["changes"])



"""
Encapsulate the different transports available to Salt.

This includes client side transport, for the ReqServer and the Publisher
"""


import logging

from salt.utils.asynchronous import SyncWrapper

log = logging.getLogger(__name__)


class ReqChannel:
    """
    Factory class to create a Sync communication channels to the ReqServer
    """

    @staticmethod
    def factory(opts, **kwargs):
        # All Sync interfaces are just wrappers around the Async ones
        return SyncWrapper(
            AsyncReqChannel.factory,
            (opts,),
            kwargs,
            loop_kwarg="io_loop",
        )

    def close(self):
        """
        Close the channel
        """
        raise NotImplementedError()

    def send(self, load, tries=3, timeout=60, raw=False):
        """
        Send "load" to the master.
        """
        raise NotImplementedError()

    def crypted_transfer_decode_dictentry(
        self, load, dictkey=None, tries=3, timeout=60
    ):
        """
        Send "load" to the master in a way that the load is only readable by
        the minion and the master (not other minions etc.)
        """
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class PushChannel:
    """
    Factory class to create Sync channel for push side of push/pull IPC
    """

    @staticmethod
    def factory(opts, **kwargs):
        return SyncWrapper(
            AsyncPushChannel.factory,
            (opts,),
            kwargs,
            loop_kwarg="io_loop",
        )

    def send(self, load, tries=3, timeout=60):
        """
        Send load across IPC push
        """
        raise NotImplementedError()


class PullChannel:
    """
    Factory class to create Sync channel for pull side of push/pull IPC
    """

    @staticmethod
    def factory(opts, **kwargs):
        return SyncWrapper(
            AsyncPullChannel.factory,
            (opts,),
            kwargs,
            loop_kwarg="io_loop",
        )


# TODO: better doc strings
class AsyncChannel:
    """
    Parent class for Async communication channels
    """

    # Resolver is used by Tornado TCPClient.
    # This static field is shared between
    # AsyncReqChannel and AsyncPubChannel.
    # This will check to make sure the Resolver
    # is configured before first use.
    _resolver_configured = False

    @classmethod
    def _config_resolver(cls, num_threads=10):
        import salt.ext.tornado.netutil

        salt.ext.tornado.netutil.Resolver.configure(
            "salt.ext.tornado.netutil.ThreadedResolver", num_threads=num_threads
        )
        cls._resolver_configured = True


# TODO: better doc strings
class AsyncReqChannel(AsyncChannel):
    """
    Factory class to create a Async communication channels to the ReqServer
    """

    @classmethod
    def factory(cls, opts, **kwargs):
        # Default to ZeroMQ for now
        ttype = "zeromq"

        # determine the ttype
        if "transport" in opts:
            ttype = opts["transport"]
        elif "transport" in opts.get("pillar", {}).get("master", {}):
            ttype = opts["pillar"]["master"]["transport"]

        # switch on available ttypes
        if ttype == "zeromq":
            import salt.transport.zeromq

            return salt.transport.zeromq.AsyncZeroMQReqChannel(opts, **kwargs)
        elif ttype == "tcp":
            if not cls._resolver_configured:
                # TODO: add opt to specify number of resolver threads
                AsyncChannel._config_resolver()
            import salt.transport.tcp

            return salt.transport.tcp.AsyncTCPReqChannel(opts, **kwargs)
        elif ttype == "local":
            raise Exception("There's no AsyncLocalChannel implementation yet")
            # import salt.transport.local
            # return salt.transport.local.AsyncLocalChannel(opts, **kwargs)
        else:
            raise Exception("Channels are only defined for tcp, zeromq, and local")
            # return NewKindOfChannel(opts, **kwargs)

    def send(self, load, tries=3, timeout=60, raw=False):
        """
        Send "load" to the master.
        """
        raise NotImplementedError()

    def crypted_transfer_decode_dictentry(
        self, load, dictkey=None, tries=3, timeout=60
    ):
        """
        Send "load" to the master in a way that the load is only readable by
        the minion and the master (not other minions etc.)
        """
        raise NotImplementedError()

    def close(self):
        """
        Close the channel
        """
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class AsyncPubChannel(AsyncChannel):
    """
    Factory class to create subscription channels to the master's Publisher
    """

    @classmethod
    def factory(cls, opts, **kwargs):
        # Default to ZeroMQ for now
        ttype = "zeromq"

        # determine the ttype
        if "transport" in opts:
            ttype = opts["transport"]
        elif "transport" in opts.get("pillar", {}).get("master", {}):
            ttype = opts["pillar"]["master"]["transport"]

        # switch on available ttypes
        if ttype == "detect":
            opts["detect_mode"] = True
            log.info("Transport is set to detect; using %s", ttype)
        if ttype == "zeromq":
            import salt.transport.zeromq

            return salt.transport.zeromq.AsyncZeroMQPubChannel(opts, **kwargs)
        elif ttype == "tcp":
            if not cls._resolver_configured:
                # TODO: add opt to specify number of resolver threads
                AsyncChannel._config_resolver()
            import salt.transport.tcp

            return salt.transport.tcp.AsyncTCPPubChannel(opts, **kwargs)
        elif ttype == "local":  # TODO:
            raise Exception("There's no AsyncLocalPubChannel implementation yet")
            # import salt.transport.local
            # return salt.transport.local.AsyncLocalPubChannel(opts, **kwargs)
        else:
            raise Exception("Channels are only defined for tcp, zeromq, and local")
            # return NewKindOfChannel(opts, **kwargs)

    def connect(self):
        """
        Return a future which completes when connected to the remote publisher
        """
        raise NotImplementedError()

    def close(self):
        """
        Close the channel
        """
        raise NotImplementedError()

    def on_recv(self, callback):
        """
        When jobs are received pass them (decoded) to callback
        """
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class AsyncPushChannel:
    """
    Factory class to create IPC Push channels
    """

    @staticmethod
    def factory(opts, **kwargs):
        """
        If we have additional IPC transports other than UxD and TCP, add them here
        """
        # FIXME for now, just UXD
        # Obviously, this makes the factory approach pointless, but we'll extend later
        import salt.transport.ipc

        return salt.transport.ipc.IPCMessageClient(opts, **kwargs)


class AsyncPullChannel:
    """
    Factory class to create IPC pull channels
    """

    @staticmethod
    def factory(opts, **kwargs):
        """
        If we have additional IPC transports other than UXD and TCP, add them here
        """
        import salt.transport.ipc

        return salt.transport.ipc.IPCMessageServer(opts, **kwargs)


## Additional IPC messaging patterns should provide interfaces here, ala router/dealer, pub/sub, etc

# EOF



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import salt.modules.hadoop as hadoop
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class HadoopTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.hadoop
    """

    def setup_loader_modules(self):
        return {hadoop: {}}

    def test_version(self):
        """
        Test for Return version from hadoop version
        """
        mock = MagicMock(return_value="A \nB \n")
        with patch.dict(hadoop.__salt__, {"cmd.run": mock}):
            self.assertEqual(hadoop.version(), "B")

    def test_dfs(self):
        """
        Test for Execute a command on DFS
        """
        with patch.object(hadoop, "_hadoop_cmd", return_value="A"):
            self.assertEqual(hadoop.dfs("command"), "A")

        self.assertEqual(hadoop.dfs(), "Error: command must be provided")

    def test_dfs_present(self):
        """
        Test for Check if a file or directory is present on the distributed FS.
        """
        with patch.object(
            hadoop, "_hadoop_cmd", side_effect=["No such file or directory", "A"]
        ):
            self.assertFalse(hadoop.dfs_present("path"))
            self.assertTrue(hadoop.dfs_present("path"))

    def test_dfs_absent(self):
        """
        Test for Check if a file or directory is absent on the distributed FS.
        """
        with patch.object(
            hadoop, "_hadoop_cmd", side_effect=["No such file or directory", "A"]
        ):
            self.assertTrue(hadoop.dfs_absent("path"))
            self.assertFalse(hadoop.dfs_absent("path"))

    def test_namenode_format(self):
        """
        Test for Format a name node
        """
        with patch.object(hadoop, "_hadoop_cmd", return_value="A"):
            self.assertEqual(hadoop.namenode_format("force"), "A")



"""
Functions which implement running reactor jobs
"""
import fnmatch
import glob
import logging
import os

import salt.client
import salt.defaults.exitcodes
import salt.runner
import salt.state
import salt.utils.args
import salt.utils.cache
import salt.utils.data
import salt.utils.event
import salt.utils.files
import salt.utils.master
import salt.utils.process
import salt.utils.yaml
import salt.wheel

log = logging.getLogger(__name__)

REACTOR_INTERNAL_KEYWORDS = frozenset(
    ["__id__", "__sls__", "name", "order", "fun", "key", "state"]
)


class Reactor(salt.utils.process.SignalHandlingProcess, salt.state.Compiler):
    """
    Read in the reactor configuration variable and compare it to events
    processed on the master.
    The reactor has the capability to execute pre-programmed executions
    as reactions to events
    """

    aliases = {
        "cmd": "local",
    }

    def __init__(self, opts, **kwargs):
        super().__init__(**kwargs)
        local_minion_opts = opts.copy()
        local_minion_opts["file_client"] = "local"
        self.minion = salt.minion.MasterMinion(local_minion_opts)
        salt.state.Compiler.__init__(self, opts, self.minion.rend)
        self.is_leader = True

    def render_reaction(self, glob_ref, tag, data):
        """
        Execute the render system against a single reaction file and return
        the data structure
        """
        react = {}

        if glob_ref.startswith("salt://"):
            glob_ref = self.minion.functions["cp.cache_file"](glob_ref) or ""
        globbed_ref = glob.glob(glob_ref)
        if not globbed_ref:
            log.error(
                "Can not render SLS %s for tag %s. File missing or not found.",
                glob_ref,
                tag,
            )
        for fn_ in globbed_ref:
            try:
                res = self.render_template(fn_, tag=tag, data=data)

                # for #20841, inject the sls name here since verify_high()
                # assumes it exists in case there are any errors
                for name in res:
                    res[name]["__sls__"] = fn_

                react.update(res)
            except Exception:  # pylint: disable=broad-except
                log.exception('Failed to render "%s": ', fn_)
        return react

    def list_reactors(self, tag):
        """
        Take in the tag from an event and return a list of the reactors to
        process
        """
        log.debug("Gathering reactors for tag %s", tag)
        reactors = []
        if isinstance(self.opts["reactor"], str):
            try:
                with salt.utils.files.fopen(self.opts["reactor"]) as fp_:
                    react_map = salt.utils.yaml.safe_load(fp_)
            except OSError:
                log.error('Failed to read reactor map: "%s"', self.opts["reactor"])
            except Exception:  # pylint: disable=broad-except
                log.error(
                    'Failed to parse YAML in reactor map: "%s"', self.opts["reactor"]
                )
        else:
            react_map = self.opts["reactor"]
        for ropt in react_map:
            if not isinstance(ropt, dict):
                continue
            if len(ropt) != 1:
                continue
            key = next(iter(ropt.keys()))
            val = ropt[key]
            if fnmatch.fnmatch(tag, key):
                if isinstance(val, str):
                    reactors.append(val)
                elif isinstance(val, list):
                    reactors.extend(val)
        return reactors

    def list_all(self):
        """
        Return a list of the reactors
        """
        if isinstance(self.minion.opts["reactor"], str):
            log.debug("Reading reactors from yaml %s", self.opts["reactor"])
            try:
                with salt.utils.files.fopen(self.opts["reactor"]) as fp_:
                    react_map = salt.utils.yaml.safe_load(fp_)
            except OSError:
                log.error('Failed to read reactor map: "%s"', self.opts["reactor"])
            except Exception:  # pylint: disable=broad-except
                log.error(
                    'Failed to parse YAML in reactor map: "%s"', self.opts["reactor"]
                )
        else:
            log.debug("Not reading reactors from yaml")
            react_map = self.minion.opts["reactor"]
        return react_map

    def add_reactor(self, tag, reaction):
        """
        Add a reactor
        """
        reactors = self.list_all()
        for reactor in reactors:
            _tag = next(iter(reactor.keys()))
            if _tag == tag:
                return {"status": False, "comment": "Reactor already exists."}

        self.minion.opts["reactor"].append({tag: reaction})
        return {"status": True, "comment": "Reactor added."}

    def delete_reactor(self, tag):
        """
        Delete a reactor
        """
        reactors = self.list_all()
        for reactor in reactors:
            _tag = next(iter(reactor.keys()))
            if _tag == tag:
                self.minion.opts["reactor"].remove(reactor)
                return {"status": True, "comment": "Reactor deleted."}

        return {"status": False, "comment": "Reactor does not exists."}

    def resolve_aliases(self, chunks):
        """
        Preserve backward compatibility by rewriting the 'state' key in the low
        chunks if it is using a legacy type.
        """
        for idx, _ in enumerate(chunks):
            new_state = self.aliases.get(chunks[idx]["state"])
            if new_state is not None:
                chunks[idx]["state"] = new_state

    def reactions(self, tag, data, reactors):
        """
        Render a list of reactor files and returns a reaction struct
        """
        log.debug("Compiling reactions for tag %s", tag)
        high = {}
        chunks = []
        try:
            for fn_ in reactors:
                high.update(self.render_reaction(fn_, tag, data))
            if high:
                errors = self.verify_high(high)
                if errors:
                    log.error(
                        "Unable to render reactions for event %s due to "
                        "errors (%s) in one or more of the sls files (%s)",
                        tag,
                        errors,
                        reactors,
                    )
                    return []  # We'll return nothing since there was an error
                chunks = self.order_chunks(self.compile_high_data(high))
        except Exception as exc:  # pylint: disable=broad-except
            log.exception("Exception encountered while compiling reactions")

        self.resolve_aliases(chunks)
        return chunks

    def call_reactions(self, chunks):
        """
        Execute the reaction state
        """
        for chunk in chunks:
            self.wrap.run(chunk)

    def run(self):
        """
        Enter into the server loop
        """
        salt.utils.process.appendproctitle(self.__class__.__name__)

        if self.opts["reactor_niceness"] and not salt.utils.platform.is_windows():
            log.info("Reactor setting niceness to %i", self.opts["reactor_niceness"])
            os.nice(self.opts["reactor_niceness"])

        # instantiate some classes inside our new process
        with salt.utils.event.get_event(
            self.opts["__role"],
            self.opts["sock_dir"],
            self.opts["transport"],
            opts=self.opts,
            listen=True,
        ) as event:
            self.wrap = ReactWrap(self.opts)

            for data in event.iter_events(full=True):
                # skip all events fired by ourselves
                if data["data"].get("user") == self.wrap.event_user:
                    continue

                # NOTE: these events must contain the masters key in order to be accepted
                # see salt.runners.reactor for the requesting interface
                if "salt/reactors/manage" in data["tag"]:
                    master_key = salt.utils.master.get_master_key("root", self.opts)
                    if data["data"].get("key") != master_key:
                        log.error(
                            "received salt/reactors/manage event without matching"
                            " master_key. discarding"
                        )
                        continue
                if data["tag"].endswith("salt/reactors/manage/is_leader"):
                    event.fire_event(
                        {"result": self.is_leader}, "salt/reactors/manage/leader/value"
                    )
                if data["tag"].endswith("salt/reactors/manage/set_leader"):
                    # we only want to register events from the local master
                    if data["data"].get("id") == self.opts["id"]:
                        self.is_leader = data["data"]["value"]
                    event.fire_event(
                        {"result": self.is_leader}, "salt/reactors/manage/leader/value"
                    )
                if data["tag"].endswith("salt/reactors/manage/add"):
                    _data = data["data"]
                    res = self.add_reactor(_data["event"], _data["reactors"])
                    event.fire_event(
                        {"reactors": self.list_all(), "result": res},
                        "salt/reactors/manage/add-complete",
                    )
                elif data["tag"].endswith("salt/reactors/manage/delete"):
                    _data = data["data"]
                    res = self.delete_reactor(_data["event"])
                    event.fire_event(
                        {"reactors": self.list_all(), "result": res},
                        "salt/reactors/manage/delete-complete",
                    )
                elif data["tag"].endswith("salt/reactors/manage/list"):
                    event.fire_event(
                        {"reactors": self.list_all()},
                        "salt/reactors/manage/list-results",
                    )
                else:
                    # do not handle any reactions if not leader in cluster
                    if not self.is_leader:
                        continue
                    else:
                        reactors = self.list_reactors(data["tag"])
                        if not reactors:
                            continue
                        chunks = self.reactions(data["tag"], data["data"], reactors)
                        if chunks:
                            try:
                                self.call_reactions(chunks)
                            except SystemExit:
                                log.warning("Exit ignored by reactor")


class ReactWrap:
    """
    Wrapper that executes low data for the Reactor System
    """

    # class-wide cache of clients
    client_cache = None
    event_user = "Reactor"

    reaction_class = {
        "local": salt.client.LocalClient,
        "runner": salt.runner.RunnerClient,
        "wheel": salt.wheel.Wheel,
        "caller": salt.client.Caller,
    }

    def __init__(self, opts):
        self.opts = opts
        if ReactWrap.client_cache is None:
            ReactWrap.client_cache = salt.utils.cache.CacheDict(
                opts["reactor_refresh_interval"]
            )

        self.pool = salt.utils.process.ThreadPool(
            self.opts["reactor_worker_threads"],  # number of workers for runner/wheel
            queue_size=self.opts["reactor_worker_hwm"],  # queue size for those workers
        )

    def populate_client_cache(self, low):
        """
        Populate the client cache with an instance of the specified type
        """
        reaction_type = low["state"]
        if reaction_type not in self.client_cache:
            log.debug("Reactor is populating %s client cache", reaction_type)
            if reaction_type in ("runner", "wheel"):
                # Reaction types that run locally on the master want the full
                # opts passed.
                self.client_cache[reaction_type] = self.reaction_class[reaction_type](
                    self.opts
                )
                # The len() function will cause the module functions to load if
                # they aren't already loaded. We want to load them so that the
                # spawned threads don't need to load them. Loading in the
                # spawned threads creates race conditions such as sometimes not
                # finding the required function because another thread is in
                # the middle of loading the functions.
                len(self.client_cache[reaction_type].functions)
            else:
                # Reactions which use remote pubs only need the conf file when
                # instantiating a client instance.
                self.client_cache[reaction_type] = self.reaction_class[reaction_type](
                    self.opts["conf_file"]
                )

    def run(self, low):
        """
        Execute a reaction by invoking the proper wrapper func
        """
        self.populate_client_cache(low)
        try:
            l_fun = getattr(self, low["state"])
        except AttributeError:
            log.error("ReactWrap is missing a wrapper function for '%s'", low["state"])

        try:
            wrap_call = salt.utils.args.format_call(l_fun, low)
            args = wrap_call.get("args", ())
            kwargs = wrap_call.get("kwargs", {})
            # TODO: Setting user doesn't seem to work for actual remote pubs
            if low["state"] in ("runner", "wheel"):
                # Update called function's low data with event user to
                # segregate events fired by reactor and avoid reaction loops
                kwargs["__user__"] = self.event_user
                # Replace ``state`` kwarg which comes from high data compiler.
                # It breaks some runner functions and seems unnecessary.
                kwargs["__state__"] = kwargs.pop("state")
                # NOTE: if any additional keys are added here, they will also
                # need to be added to filter_kwargs()

            if "args" in kwargs:
                # New configuration
                reactor_args = kwargs.pop("args")
                for item in ("arg", "kwarg"):
                    if item in low:
                        log.warning(
                            "Reactor '%s' is ignoring '%s' param %s due to "
                            "presence of 'args' param. Check the Reactor System "
                            "documentation for the correct argument format.",
                            low["__id__"],
                            item,
                            low[item],
                        )
                if (
                    low["state"] == "caller"
                    and isinstance(reactor_args, list)
                    and not salt.utils.data.is_dictlist(reactor_args)
                ):
                    # Legacy 'caller' reactors were already using the 'args'
                    # param, but only supported a list of positional arguments.
                    # If low['args'] is a list but is *not* a dictlist, then
                    # this is actually using the legacy configuration. So, put
                    # the reactor args into kwarg['arg'] so that the wrapper
                    # interprets them as positional args.
                    kwargs["arg"] = reactor_args
                    kwargs["kwarg"] = {}
                else:
                    kwargs["arg"] = ()
                    kwargs["kwarg"] = reactor_args
                if not isinstance(kwargs["kwarg"], dict):
                    kwargs["kwarg"] = salt.utils.data.repack_dictlist(kwargs["kwarg"])
                    if not kwargs["kwarg"]:
                        log.error(
                            "Reactor '%s' failed to execute %s '%s': "
                            "Incorrect argument format, check the Reactor System "
                            "documentation for the correct format.",
                            low["__id__"],
                            low["state"],
                            low["fun"],
                        )
                        return
            else:
                # Legacy configuration
                react_call = {}
                if low["state"] in ("runner", "wheel"):
                    if "arg" not in kwargs or "kwarg" not in kwargs:
                        # Runner/wheel execute on the master, so we can use
                        # format_call to get the functions args/kwargs
                        react_fun = self.client_cache[low["state"]].functions.get(
                            low["fun"]
                        )
                        if react_fun is None:
                            log.error(
                                "Reactor '%s' failed to execute %s '%s': "
                                "function not available",
                                low["__id__"],
                                low["state"],
                                low["fun"],
                            )
                            return

                        react_call = salt.utils.args.format_call(
                            react_fun, low, expected_extra_kws=REACTOR_INTERNAL_KEYWORDS
                        )

                if "arg" not in kwargs:
                    kwargs["arg"] = react_call.get("args", ())
                if "kwarg" not in kwargs:
                    kwargs["kwarg"] = react_call.get("kwargs", {})

            # Execute the wrapper with the proper args/kwargs. kwargs['arg']
            # and kwargs['kwarg'] contain the positional and keyword arguments
            # that will be passed to the client interface to execute the
            # desired runner/wheel/remote-exec/etc. function.
            ret = l_fun(*args, **kwargs)

            if ret is False:
                log.error(
                    "Reactor '%s' failed  to execute %s '%s': "
                    "TaskPool queue is full!"
                    "Consider tuning reactor_worker_threads and/or"
                    " reactor_worker_hwm",
                    low["__id__"],
                    low["state"],
                    low["fun"],
                )

        except SystemExit:
            log.warning("Reactor '%s' attempted to exit. Ignored.", low["__id__"])
        except Exception:  # pylint: disable=broad-except
            log.error(
                "Reactor '%s' failed to execute %s '%s'",
                low["__id__"],
                low["state"],
                low["fun"],
                exc_info=True,
            )

    def runner(self, fun, **kwargs):
        """
        Wrap RunnerClient for executing :ref:`runner modules <all-salt.runners>`
        """
        return self.pool.fire_async(self.client_cache["runner"].low, args=(fun, kwargs))

    def wheel(self, fun, **kwargs):
        """
        Wrap Wheel to enable executing :ref:`wheel modules <all-salt.wheel>`
        """
        return self.pool.fire_async(self.client_cache["wheel"].low, args=(fun, kwargs))

    def local(self, fun, tgt, **kwargs):
        """
        Wrap LocalClient for running :ref:`execution modules <all-salt.modules>`
        """
        self.client_cache["local"].cmd_async(tgt, fun, **kwargs)

    def caller(self, fun, **kwargs):
        """
        Wrap LocalCaller to execute remote exec functions locally on the Minion
        """
        self.client_cache["caller"].cmd(fun, *kwargs["arg"], **kwargs["kwarg"])



"""
unit tests for the jobs runner
"""


import salt.minion
import salt.runners.jobs as jobs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class JobsTest(TestCase, LoaderModuleMockMixin):
    """
    Validate the jobs runner
    """

    def setup_loader_modules(self):
        return {
            jobs: {
                "__opts__": {"ext_job_cache": None, "master_job_cache": "local_cache"}
            }
        }

    def test_list_jobs_with_search_target(self):
        """
        test jobs.list_jobs runner with search_target args
        """
        mock_jobs_cache = {
            "20160524035503086853": {
                "Arguments": [],
                "Function": "test.ping",
                "StartTime": "2016, May 24 03:55:03.086853",
                "Target": "node-1-1.com",
                "Target-type": "glob",
                "User": "root",
            },
            "20160524035524895387": {
                "Arguments": [],
                "Function": "test.ping",
                "StartTime": "2016, May 24 03:55:24.895387",
                "Target": ["node-1-2.com", "node-1-1.com"],
                "Target-type": "list",
                "User": "sudo_ubuntu",
            },
        }

        def return_mock_jobs():
            return mock_jobs_cache

        class MockMasterMinion:

            returners = {"local_cache.get_jids": return_mock_jobs}

            def __init__(self, *args, **kwargs):
                pass

        returns = {
            "all": mock_jobs_cache,
            "node-1-1.com": mock_jobs_cache,
            "node-1-2.com": {
                "20160524035524895387": mock_jobs_cache["20160524035524895387"]
            },
            "non-existant": {},
        }

        with patch.object(salt.minion, "MasterMinion", MockMasterMinion):
            self.assertEqual(jobs.list_jobs(), returns["all"])

            self.assertEqual(
                jobs.list_jobs(search_target=["node-1-1*", "node-1-2*"]), returns["all"]
            )

            self.assertEqual(
                jobs.list_jobs(search_target="node-1-1.com"), returns["node-1-1.com"]
            )

            self.assertEqual(
                jobs.list_jobs(search_target="node-1-2.com"), returns["node-1-2.com"]
            )

            self.assertEqual(
                jobs.list_jobs(search_target="non-existant"), returns["non-existant"]
            )



"""
Management of Solaris Zones

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.modules.zoneadm, salt.modules.zonecfg
:platform:      solaris

.. versionadded:: 2017.7.0

Below are some examples of how to use this state.
Lets start with creating a zone and installing it.

.. code-block:: yaml

    omipkg1_configuration:
      zone.present:
        - name: omipkg1
        - brand: ipkg
        - zonepath: /zones/omipkg1
        - properties:
          - autoboot: true
          - ip-type: exclusive
          - cpu-shares: 50
        - resources:
          - attr:
            - name: owner
            - value: Jorge Schrauwen
            - type: string
          - attr:
            - name: description
            - value: OmniOS ipkg zone for testing
            - type: string
          - capped-memory:
            - physical: 64M
    omipkg1_installation:
      zone.installed:
        - name: omipkg1
        - require:
            - zone: omipkg1_configuration
    omipkg1_running:
      zone.booted:
        - name: omipkg1
        - require:
            - zone: omipkg1_installation

A zone without network access is not very useful. We could update
the zone.present state in the example above to add a network interface
or we could use a separate state for this.

.. code-block:: yaml

    omipkg1_network:
      zone.resource_present:
        - name: omipkg1
        - resource_type: net
        - resource_selector_property: mac-addr
        - resource_selector_value: "02:08:20:a2:a3:10"
        - physical: znic1
        - require:
            - zone: omipkg1_configuration

Since this is a single tenant system having the owner attribute is pointless.
Let's remove that attribute.

.. note::
    The following state run the omipkg1_configuration state will add it again!
    If the entire configuration is managed it would be better to add resource_prune
    and optionally the resource_selector_property properties to the resource.

.. code-block:: yaml

    omipkg1_strip_owner:
      zone.resource_present:
        - name: omipkg1
        - resource_type: attr
        - resource_selector_property: name
        - resource_selector_value: owner
        - require:
            - zone: omipkg1_configuration

Let's bump the zone's CPU shares a bit.

.. note::
    The following state run the omipkg1_configuration state will set it to 50 again.
    Update the entire zone configuration is managed you should update it there instead.

.. code-block:: yaml

    omipkg1_more_cpu:
      zone.property_present:
        - name: omipkg1
        - property: cpu-shares
        - value: 100

Or we can remove the limit altogether!

.. note::
    The following state run the omipkg1_configuration state will set it to 50 again.
    Update the entire zone configuration is managed you should set the
    property to None (nothing after the :) instead.

.. code-block:: yaml

    omipkg1_no_cpu:
      zone.property_absent:
        - name: omipkg1
        - property: cpu-shares

"""

import logging

import salt.utils.args
import salt.utils.atomicfile
import salt.utils.files
from salt.exceptions import CommandExecutionError
from salt.modules.zonecfg import _parse_value, _zonecfg_resource_default_selectors
from salt.utils.dictupdate import merge as merge_dict
from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)

__func_alias__ = {
    "import_": "import",
}

# Define the state's virtual name
__virtualname__ = "zone"


def __virtual__():
    """
    Provides zone state on Solaris
    """
    if "zonecfg.create" in __salt__ and "zoneadm.install" in __salt__:
        return True
    else:
        return (
            False,
            "{} state module can only be loaded on Solaris platforms".format(
                __virtualname__
            ),
        )


def property_present(name, property, value):
    """
    Ensure property has a certain value

    name : string
        name of the zone
    property : string
        name of property
    value : string
        value of property

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    ## sanitize input
    value = _parse_value(value)

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        ## zone exists
        zonecfg = __salt__["zonecfg.info"](name, show_all=True)
        if property not in zonecfg or zonecfg[property] != _parse_value(value):
            if __opts__["test"]:
                ret["result"] = True
            else:
                # update property
                zonecfg_res = __salt__["zonecfg.set_property"](name, property, value)
                ret["result"] = zonecfg_res["status"]
                if "messages" in zonecfg_res:
                    ret["comment"] = zonecfg_res["message"]
            if ret["result"]:
                ret["changes"][property] = _parse_value(value)
                if ret["comment"] == "":
                    ret["comment"] = "The property {} is was updated to {}.".format(
                        property, value
                    )
            elif ret["comment"] == "":
                if ret["comment"] == "":
                    ret["comment"] = "The property {} is was not updated to {}!".format(
                        property, value
                    )
        else:
            ret["result"] = True
            ret["comment"] = "The property {} is already set to {}.".format(
                property, value
            )
    else:
        ## zone does not exist
        ret["result"] = False
        ret[
            "comment"
        ] = "The zone {} is not in the configured, installed, or booted state.".format(
            name
        )

    return ret


def property_absent(name, property):
    """
    Ensure property is absent

    name : string
        name of the zone
    property : string
        name of property

    .. note::
        This does a zoneacfg clear call. So the property may be reset to a default value!
        Does has the side effect of always having to be called.

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        ## zone exists
        zonecfg = __salt__["zonecfg.info"](name, show_all=True)
        if property in zonecfg:
            if __opts__["test"]:
                ret["result"] = True
            else:
                # clear property
                zonecfg_res = __salt__["zonecfg.clear_property"](name, property)
                zonecfg_new = __salt__["zonecfg.info"](name, show_all=True)
                ret["result"] = zonecfg_res["status"]
                if "messages" in zonecfg_res:
                    ret["comment"] = zonecfg_res["message"]
            if ret["result"]:
                if property not in zonecfg_new:
                    ret["changes"][property] = None
                elif zonecfg[property] != zonecfg_new[property]:
                    ret["changes"][property] = zonecfg_new[property]
                if ret["comment"] == "":
                    ret["comment"] = "The property {} was cleared!".format(property)
            elif ret["comment"] == "":
                if ret["comment"] == "":
                    ret["comment"] = "The property {} did not get cleared!".format(
                        property
                    )
        else:
            ret["result"] = True
            ret["comment"] = "The property {} does not exist!".format(property)
    else:
        ## zone does not exist
        ret["result"] = False
        ret[
            "comment"
        ] = "The zone {} is not in the configured, installed, or booted state.".format(
            name
        )

    return ret


def resource_present(
    name, resource_type, resource_selector_property, resource_selector_value, **kwargs
):
    """
    Ensure resource exists with provided properties

    name : string
        name of the zone
    resource_type : string
        type of resource
    resource_selector_property : string
        unique resource identifier
    resource_selector_value : string
        value for resource selection
    kwargs : string|int|...
        resource properties

    .. warning::
        Both resource_selector_property and resource_selector_value must be
        provided, some properties like ``name`` are already reserved by salt in
        states.

    .. note::
        You can set both resource_selector_property and resource_selector_value
        to None for resources that do not require them.

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # sanitize input
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    resource_selector_value = _parse_value(resource_selector_value)
    for k, v in kwargs.items():
        kwargs[k] = _parse_value(kwargs[k])

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        ## zone exists
        zonecfg = __salt__["zonecfg.info"](name, show_all=True)

        ## update kwargs
        zonecfg_kwargs = {}
        zonecfg_kwargs.update(kwargs)
        zonecfg_kwargs["zone"] = name
        zonecfg_kwargs["resource_type"] = resource_type
        zonecfg_kwargs["resource_selector"] = resource_selector_property
        if resource_selector_property:
            zonecfg_kwargs[resource_selector_property] = resource_selector_value

        ## check update or add
        if resource_type in zonecfg:
            for resource in zonecfg[resource_type]:
                if (
                    not resource_selector_property
                    or resource[resource_selector_property] == resource_selector_value
                ):
                    ret["result"] = True
                    if resource_selector_property:
                        ret["comment"] = "the {} resource {} is up to date.".format(
                            resource_type,
                            resource_selector_value,
                        )
                    else:
                        ret["comment"] = "the {} resource is up to date.".format(
                            resource_type,
                        )

                    ## check if update reauired
                    for key in kwargs:
                        log.debug(
                            "zone.resource_preent - key=%s value=%s current_value=%s",
                            key,
                            resource[key] if key in resource else None,
                            _parse_value(kwargs[key]),
                        )
                        # note: something odd with ncpus property, we fix it here for now
                        if key == "ncpus" and key in kwargs:
                            kwargs[key] = "{:.2f}".format(float(kwargs[key]))

                        if key not in resource:
                            ret["result"] = None
                        elif resource[key] != _parse_value(kwargs[key]):
                            ret["result"] = None

                    ## do update
                    if ret["result"] is None:
                        if __opts__["test"]:
                            ret["result"] = True
                        else:
                            ## update resource
                            zonecfg_res = __salt__["zonecfg.update_resource"](
                                **zonecfg_kwargs
                            )
                            ret["result"] = zonecfg_res["status"]
                            if "message" in zonecfg_res:
                                ret["comment"] = zonecfg_res["message"]

                        if ret["result"]:
                            ret["changes"][resource_type] = {}
                            if resource_selector_property:
                                ret["changes"][resource_type][
                                    resource_selector_value
                                ] = {}
                            for key in kwargs if ret["result"] else []:
                                if resource_selector_property:
                                    ret["changes"][resource_type][
                                        resource_selector_value
                                    ][key] = _parse_value(kwargs[key])
                                else:
                                    ret["changes"][resource_type][key] = _parse_value(
                                        kwargs[key]
                                    )
                            if ret["comment"] == "":
                                if resource_selector_property:
                                    ret[
                                        "comment"
                                    ] = "The {} resource {} was updated.".format(
                                        resource_type,
                                        resource_selector_value,
                                    )
                                else:
                                    ret[
                                        "comment"
                                    ] = "The {} resource was updated.".format(
                                        resource_type,
                                    )
                        elif ret["comment"] == "":
                            if resource_selector_property:
                                ret[
                                    "comment"
                                ] = "The {} resource {} was not updated.".format(
                                    resource_type,
                                    resource_selector_value,
                                )
                            else:
                                ret[
                                    "comment"
                                ] = "The {} resource was not updated.".format(
                                    resource_type,
                                )
        if ret["result"] is None:
            ## add
            if __opts__["test"]:
                ret["result"] = True
            else:
                ## add resource
                if "resource_selector" in zonecfg_kwargs:
                    del zonecfg_kwargs["resource_selector"]
                zonecfg_res = __salt__["zonecfg.add_resource"](**zonecfg_kwargs)
                ret["result"] = zonecfg_res["status"]
                if "message" in zonecfg_res:
                    ret["comment"] = zonecfg_res["message"]

            if ret["result"]:
                ret["changes"][resource_type] = {}
                if resource_selector_property:
                    ret["changes"][resource_type][resource_selector_value] = {}
                for key in kwargs if ret["result"] else []:
                    if resource_selector_property:
                        ret["changes"][resource_type][resource_selector_value][
                            key
                        ] = _parse_value(kwargs[key])
                    else:
                        ret["changes"][resource_type][key] = _parse_value(kwargs[key])
                if ret["comment"] == "":
                    ret["comment"] = "The {} resource {} was added.".format(
                        resource_type,
                        resource_selector_value,
                    )
            elif ret["comment"] == "":
                ret["comment"] = "The {} resource {} was not added.".format(
                    resource_type,
                    resource_selector_value,
                )
    else:
        ## zone does not exist
        ret["result"] = False
        ret[
            "comment"
        ] = "The zone {} is not in the configured, installed, or booted state.".format(
            name
        )

    return ret


def resource_absent(
    name, resource_type, resource_selector_property, resource_selector_value
):
    """
    Ensure resource is absent

    name : string
        name of the zone
    resource_type : string
        type of resource
    resource_selector_property : string
        unique resource identifier
    resource_selector_value : string
        value for resource selection

    .. warning::
        Both resource_selector_property and resource_selector_value must be provided, some properties
        like ```name``` are already reserved by salt in there states.

    .. note::
        You can set both resource_selector_property and resource_selector_value to None for
        resources that do not require them.

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # sanitize input
    if resource_selector_property:
        resource_selector_value = _parse_value(resource_selector_value)
    else:
        resource_selector_value = None

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        ## zone exists
        zonecfg = __salt__["zonecfg.info"](name, show_all=True)
        if resource_type in zonecfg:
            for resource in zonecfg[resource_type]:
                if __opts__["test"]:
                    ret["result"] = True
                elif not resource_selector_property:
                    zonecfg_res = __salt__["zonecfg.remove_resource"](
                        zone=name,
                        resource_type=resource_type,
                        resource_key=None,
                        resource_value=None,
                    )
                    ret["result"] = zonecfg_res["status"]
                    if zonecfg_res["status"]:
                        ret["changes"][resource_type] = "removed"
                        if ret["comment"] == "":
                            ret["comment"] = "The {} resource was removed.".format(
                                resource_type,
                            )
                    elif "messages" in zonecfg_res:
                        ret["comment"] = zonecfg_res["message"]
                    else:
                        ret["comment"] = "The {} resource was not removed.".format(
                            resource_type,
                        )
                elif resource[resource_selector_property] == resource_selector_value:
                    zonecfg_res = __salt__["zonecfg.remove_resource"](
                        zone=name,
                        resource_type=resource_type,
                        resource_key=resource_selector_property,
                        resource_value=resource_selector_value,
                    )
                    ret["result"] = zonecfg_res["status"]
                    if zonecfg_res["status"]:
                        ret["changes"][resource_type] = {}
                        ret["changes"][resource_type][
                            resource_selector_value
                        ] = "removed"
                        if ret["comment"] == "":
                            ret["comment"] = "The {} resource {} was removed.".format(
                                resource_type,
                                resource_selector_value,
                            )
                    elif "messages" in zonecfg_res:
                        ret["comment"] = zonecfg_res["message"]
                    else:
                        ret["comment"] = "The {} resource {} was not removed.".format(
                            resource_type,
                            resource_selector_value,
                        )

        # resource already absent
        if ret["result"] is None:
            ret["result"] = True
            ret["comment"] = "The {} resource {} was absent.".format(
                resource_type,
                resource_selector_value,
            )
    else:
        ## zone does not exist
        ret["result"] = False
        ret[
            "comment"
        ] = "The zone {} is not in the configured, installed, or booted state.".format(
            name
        )

    return ret


def booted(name, single=False):
    """
    Ensure zone is booted

    name : string
        name of the zone
    single : boolean
        boot in single usermode

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True)
    if name in zones:
        ## zone exists
        if zones[name]["state"] == "running":
            ## zone is running
            ret["result"] = True
            ret["comment"] = "Zone {} already booted".format(name)
        else:
            ## try and boot the zone
            if not __opts__["test"]:
                zoneadm_res = __salt__["zoneadm.boot"](name, single)
            if __opts__["test"] or zoneadm_res["status"]:
                ret["result"] = True
                ret["changes"][name] = "booted"
                ret["comment"] = "Zone {} booted".format(name)
            else:
                ret["result"] = False
                ret["comment"] = "Failed to boot {}".format(name)
    else:
        ## zone does not exist
        ret["comment"] = []
        ret["comment"].append(
            "The zone {} is not in the installed or booted state.".format(name)
        )
        for zone in zones:
            if zones[zone]["uuid"] == name:
                ret["comment"].append(
                    "The zone {} has a uuid of {}, please use the zone name instead!".format(
                        zone,
                        name,
                    )
                )

        ret["result"] = False
        ret["comment"] = "\n".join(ret["comment"])

    return ret


def halted(name, graceful=True):
    """
    Ensure zone is halted

    name : string
        name of the zone
    graceful : boolean
        use shutdown instead of halt if true

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True)
    if name in zones:
        ## zone exists
        if zones[name]["state"] != "running":
            ## zone is not running
            ret["result"] = True
            ret["comment"] = "Zone {} already halted".format(name)
        else:
            ## try and halt the zone
            if not __opts__["test"]:
                zoneadm_res = (
                    __salt__["zoneadm.shutdown"](name)
                    if graceful
                    else __salt__["zoneadm.halt"](name)
                )
            if __opts__["test"] or zoneadm_res["status"]:
                ret["result"] = True
                ret["changes"][name] = "halted"
                ret["comment"] = "Zone {} halted".format(name)
            else:
                ret["result"] = False
                ret["comment"] = "Failed to halt {}".format(name)
    else:
        ## zone does not exist
        ret["comment"] = []
        ret["comment"].append("The zone {} is not in the installed state.".format(name))
        for zone in zones:
            if zones[zone]["uuid"] == name:
                ret["comment"].append(
                    "The zone {} has a uuid of {}, please use the zone name instead!".format(
                        zone,
                        name,
                    )
                )
        ## note: a non existing zone is not running, we do not consider this a failure
        ret["result"] = True
        ret["comment"] = "\n".join(ret["comment"])

    return ret


def export(name, path, replace=False):
    """
    Export a zones configuration

    name : string
        name of the zone
    path : string
        path of file to export too.
    replace : boolean
        replace the file if it exists

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        ## zone exists
        if __opts__["test"]:
            ## pretend we did the correct thing
            ret["result"] = True
            ret["comment"] = "Zone configartion for {} exported to {}".format(
                name,
                path,
            )
            ret["changes"][name] = "exported"
            if __salt__["file.file_exists"](path) and not replace:
                ret["result"] = False
                ret["changes"] = {}
                ret[
                    "comment"
                ] = "File {} exists, zone configuration for {} not exported.".format(
                    path,
                    name,
                )
        else:
            ## export and update file
            cfg_tmp = salt.utils.files.mkstemp()
            __salt__["zonecfg.export"](name, cfg_tmp)
            if not __salt__["file.file_exists"](path):
                ## move cfg_tmp to path
                try:
                    __salt__["file.move"](cfg_tmp, path)
                except CommandExecutionError:
                    if __salt__["file.file_exists"](cfg_tmp):
                        __salt__["file.remove"](cfg_tmp)
                    ret["result"] = False
                    ret[
                        "comment"
                    ] = "Unable to export zone configuration for {} to {}!".format(
                        name,
                        path,
                    )
                else:
                    ret["result"] = True
                    ret[
                        "comment"
                    ] = "Zone configuration for {} was exported to {}.".format(
                        name,
                        path,
                    )
                    ret["changes"][name] = "exported"
            else:
                cfg_diff = __salt__["file.get_diff"](path, cfg_tmp)
                if not cfg_diff:
                    ret["result"] = True
                    ret[
                        "comment"
                    ] = "Zone configuration for {} was already exported to {}.".format(
                        name, path
                    )
                    if __salt__["file.file_exists"](cfg_tmp):
                        __salt__["file.remove"](cfg_tmp)
                else:
                    if replace:
                        try:
                            __salt__["file.move"](cfg_tmp, path)
                        except CommandExecutionError:
                            if __salt__["file.file_exists"](cfg_tmp):
                                __salt__["file.remove"](cfg_tmp)
                            ret["result"] = False
                            ret[
                                "comment"
                            ] = "Unable to be re-export zone configuration for {}" " to {}!".format(
                                name,
                                path,
                            )
                        else:
                            ret["result"] = True
                            ret[
                                "comment"
                            ] = "Zone configuration for {} was re-exported to {}.".format(
                                name,
                                path,
                            )
                            ret["changes"][name] = "exported"
                    else:
                        ret["result"] = False
                        ret["comment"] = (
                            "Zone configuration for {} is different from the one"
                            " exported to {}!".format(name, path)
                        )
                        if __salt__["file.file_exists"](cfg_tmp):
                            __salt__["file.remove"](cfg_tmp)
    else:
        ## zone does not exist
        ret["comment"] = []
        ret["comment"].append("The zone {} does not exist.".format(name))
        for zone in zones:
            if zones[zone]["uuid"] == name:
                ret["comment"].append(
                    "The zone {} has a uuid of {}, please use the zone name instead!".format(
                        name,
                        path,
                    )
                )

        ret["result"] = False
        ret["comment"] = "\n".join(ret["comment"])

    return ret


def import_(name, path, mode="import", nodataset=False, brand_opts=None):
    """
    Import a zones configuration

    name : string
        name of the zone
    path : string
        path of the configuration file to import
    mode : string
        either import, install, or attach
    nodataset : boolean
        do not create a ZFS file system
    brand_opts : boolean
        brand specific options to pass

    .. note::
        The mode argument can be set to ``import``, ``install``, or ``attach``.
        ``import``: will only import the configuration
        ``install``: will import and then try to install the zone
        ``attach``: will import and then try to attach of the zone

    .. code-block:: yaml

        omipkg1:
          zone.import:
            - path: /foo/bar/baz
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name not in zones:
        if __opts__["test"]:
            ret["result"] = True
            ret["comment"] = "Zone {} was imported from {}.".format(
                name,
                path,
            )
            ret["changes"][name] = "imported"
        else:
            if __salt__["file.file_exists"](path):
                res_import = __salt__["zonecfg.import"](name, path)
                if not res_import["status"]:
                    ret["result"] = False
                    ret[
                        "comment"
                    ] = "Unable to import zone configuration for {}!".format(name)
                else:
                    ret["result"] = True
                    ret["changes"][name] = "imported"
                    ret["comment"] = "Zone {} was imported from {}.".format(
                        name,
                        path,
                    )
                    if mode.lower() == "attach":
                        res_attach = __salt__["zoneadm.attach"](name, False, brand_opts)
                        ret["result"] = res_attach["status"]
                        if res_attach["status"]:
                            ret["changes"][name] = "attached"
                            ret["comment"] = "Zone {} was attached from {}.".format(
                                name,
                                path,
                            )
                        else:
                            ret["comment"] = []
                            ret["comment"].append(
                                "Failed to attach zone {} from {}!".format(
                                    name,
                                    path,
                                )
                            )
                            if "message" in res_attach:
                                ret["comment"].append(res_attach["message"])
                            ret["comment"] = "\n".join(ret["comment"])
                    if mode.lower() == "install":
                        res_install = __salt__["zoneadm.install"](
                            name, nodataset, brand_opts
                        )
                        ret["result"] = res_install["status"]
                        if res_install["status"]:
                            ret["changes"][name] = "installed"
                            ret["comment"] = "Zone {} was installed from {}.".format(
                                name,
                                path,
                            )
                        else:
                            ret["comment"] = []
                            ret["comment"].append(
                                "Failed to install zone {} from {}!".format(
                                    name,
                                    path,
                                )
                            )
                            if "message" in res_install:
                                ret["comment"].append(res_install["message"])
                            ret["comment"] = "\n".join(ret["comment"])
            else:
                ret["result"] = False
                ret[
                    "comment"
                ] = "The file {} does not exists, unable to import!".format(path)
    else:
        ## zone exist
        ret["result"] = True
        ret["comment"] = "Zone {} already exists, not importing configuration.".format(
            name
        )

    return ret


def present(name, brand, zonepath, properties=None, resources=None):
    """
    Ensure a zone with certain properties and resources

    name : string
        name of the zone
    brand : string
        brand of the zone
    zonepath : string
        path of the zone
    properties : list of key-value pairs
        dict of properties
    resources : list of key-value pairs
        dict of resources

    .. note::
        If the zone does not exist it will not be installed.
        You can use the ```zone.installed``` state for this.

    .. note::
        Default resource selectors:
            - fs: dir
            - net: mac-addr
            - device: match
            - rctl: name
            - attr: name
            - dataset: name
            - admin: user

    .. warning::
        Properties and resource will not be removed when they
        are absent from the state!

        For properties, simple set them to ```None```.

        For resources, add the ```resource_prune``` property
        and set it to ```True```. Also specify the
        ```resource_selector_property``` if the default is not
        the one you want.

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": []}

    ## sanitize defaults
    if not properties:
        properties = []
    if not resources:
        resources = []
    properties.append(OrderedDict({"brand": brand}))
    properties.append(OrderedDict({"zonepath": zonepath}))

    zones = __salt__["zoneadm.list"](installed=True, configured=True)

    ## test mode only has limited support
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"].append(
            "Cannot determine of changes would happen to the zone {}.".format(name)
        )

    ## create zone if needed
    if name not in zones:
        if __opts__["test"]:
            ## we pretend we created the zone
            res_create = {"status": True}
            ret["comment"] = []
        else:
            ## create and install
            res_create = __salt__["zonecfg.create"](name, brand, zonepath)
        if res_create["status"]:
            ret["result"] = True
            ret["changes"][name] = "created"
            ret["comment"].append("The zone {} was created.".format(name))

    if not __opts__["test"]:
        ret["result"] = True
        if isinstance(properties, list):
            for prop in properties:
                if not isinstance(prop, OrderedDict) or len(prop) != 1:
                    log.warning("zone.present - failed to parse property: %s", prop)
                    continue
                for key, value in prop.items():
                    res = None
                    if not value:
                        res = property_absent(name, key)
                    elif value:
                        res = property_present(name, key, value)
                    if res:
                        ret["result"] = ret["result"] if res["result"] else False
                        ret["comment"].append(res["comment"])
                        if len(res["changes"]) > 0:
                            if "property" not in ret["changes"]:
                                ret["changes"]["property"] = {}
                            ret["changes"]["property"] = merge_dict(
                                ret["changes"]["property"], res["changes"]
                            )
        if isinstance(resources, list):
            for resource in resources:
                if not isinstance(prop, OrderedDict) or len(prop) != 1:
                    log.warning("zone.present - failed to parse resource: %s", resource)
                    continue
                for key, value in resource.items():
                    zonecfg = __salt__["zonecfg.info"](name, show_all=True)
                    resource_cfg = {}
                    resource_cfg["resource_type"] = key
                    if isinstance(value, list):
                        for respv in value:
                            resource_cfg.update(dict(respv))

                    resource_prune = False
                    resource_selector_property = None
                    if "resource_prune" in resource_cfg:
                        resource_prune = resource_cfg["resource_prune"]
                        del resource_cfg["resource_prune"]
                    if "resource_selector_property" in resource_cfg:
                        resource_selector_property = resource_cfg[
                            "resource_selector_property"
                        ]
                        del resource_cfg["resource_selector_property"]
                    if (
                        not resource_selector_property
                        and key in _zonecfg_resource_default_selectors
                    ):
                        resource_selector_property = (
                            _zonecfg_resource_default_selectors[key]
                        )

                    res = None
                    if resource_prune:
                        res = resource_absent(
                            name,
                            resource_cfg["resource_type"],
                            resource_selector_property=resource_selector_property,
                            resource_selector_value=resource_cfg[
                                resource_selector_property
                            ]
                            if resource_selector_property
                            else None,
                        )
                    else:
                        resource_cfg[
                            "resource_selector_property"
                        ] = resource_selector_property
                        if resource_selector_property in resource_cfg:
                            resource_cfg["resource_selector_value"] = resource_cfg[
                                resource_selector_property
                            ]
                        else:
                            resource_cfg["resource_selector_value"] = None
                        resource_cfg[
                            "name"
                        ] = name  # we do this last because name can also be a attrib value
                        res = resource_present(**resource_cfg)
                    if res:
                        ret["result"] = ret["result"] if res["result"] else False
                        ret["comment"].append(res["comment"])
                        if len(res["changes"]) > 0:
                            if "resource" not in ret["changes"]:
                                ret["changes"]["resource"] = {}
                            ret["changes"]["resource"] = merge_dict(
                                ret["changes"]["resource"], res["changes"]
                            )

    if isinstance(ret["comment"], list):
        ret["comment"] = "\n".join(ret["comment"])

    return ret


def absent(name, uninstall=False):
    """
    Ensure a zone is absent

    name : string
        name of the zone
    uninstall : boolean
        when true, uninstall instead of detaching the zone first.

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        if __opts__["test"]:
            ret["result"] = True
            ret["changes"][name] = "removed"
            ret["comment"] = "Zone {} was removed.".format(name)
        else:
            ret["result"] = True
            if uninstall and zones[name]["state"] in ["running", "installed"]:
                res_halt = __salt__["zoneadm.halt"](name)
                res_uninstall = __salt__["zoneadm.uninstall"](name)
                ret["result"] = res_uninstall["status"]
                if ret["result"]:
                    ret["changes"][name] = "uninstalled"
                    ret["comment"] = "The zone {} was uninstalled.".format(name)
                else:
                    ret["comment"] = []
                    ret["comment"].append("Failed to uninstall zone {}!".format(name))
                    if "message" in res_uninstall:
                        ret["comment"].append(res_uninstall["message"])
                    ret["comment"] = "\n".join(ret["comment"])
            elif zones[name]["state"] == "installed":
                res_detach = __salt__["zoneadm.detach"](name)
                ret["result"] = res_detach["status"]
                if ret["result"]:
                    ret["changes"][name] = "detached"
                    ret["comment"] = "The zone {} was detached.".format(name)
                else:
                    ret["comment"] = []
                    ret["comment"].append("Failed to detach zone {}!".format(name))
                    if "message" in res_detach:
                        ret["comment"].append(res_detach["message"])
                    ret["comment"] = "\n".join(ret["comment"])
            if ret["result"]:
                res_delete = __salt__["zonecfg.delete"](name)
                ret["result"] = res_delete["status"]
                if ret["result"]:
                    ret["changes"][name] = "deleted"
                    ret["comment"] = "The zone {} was delete.".format(name)
                else:
                    ret["comment"] = []
                    ret["comment"].append("Failed to delete zone {}!".format(name))
                    if "message" in res_delete:
                        ret["comment"].append(res_delete["message"])
                    ret["comment"] = "\n".join(ret["comment"])
    else:
        ret["result"] = True
        ret["comment"] = "Zone {} does not exist.".format(name)

    return ret


def attached(name, force=False):
    """
    Ensure zone is attached

    name : string
        name of the zone
    force : boolean
        force attach the zone

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        if zones[name]["state"] == "configured":
            if __opts__["test"]:
                res_attach = {"status": True}
            else:
                res_attach = __salt__["zoneadm.attach"](name, force)
            ret["result"] = res_attach["status"]
            if ret["result"]:
                ret["changes"][name] = "attached"
                ret["comment"] = "The zone {} was attached.".format(name)
            else:
                ret["comment"] = []
                ret["comment"].append("Failed to attach zone {}!".format(name))
                if "message" in res_attach:
                    ret["comment"].append(res_attach["message"])
                ret["comment"] = "\n".join(ret["comment"])
        else:
            ret["result"] = True
            ret["comment"] = "zone {} already attached.".format(name)
    else:
        ret["result"] = False
        ret["comment"] = "zone {} is not configured!".format(name)

    return ret


def detached(name):
    """
    Ensure zone is detached

    name : string
        name of the zone

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        if zones[name]["state"] != "configured":
            if __opts__["test"]:
                res_detach = {"status": True}
            else:
                res_detach = __salt__["zoneadm.detach"](name)
            ret["result"] = res_detach["status"]
            if ret["result"]:
                ret["changes"][name] = "detached"
                ret["comment"] = "The zone {} was detached.".format(name)
            else:
                ret["comment"] = []
                ret["comment"].append("Failed to detach zone {}!".format(name))
                if "message" in res_detach:
                    ret["comment"].append(res_detach["message"])
                ret["comment"] = "\n".join(ret["comment"])
        else:
            ret["result"] = True
            ret["comment"] = "zone {} already detached.".format(name)
    else:
        ## note: a non existing zone is not attached, we do not consider this a failure
        ret["result"] = True
        ret["comment"] = "zone {} is not configured!".format(name)

    return ret


def installed(name, nodataset=False, brand_opts=None):
    """
    Ensure zone is installed

    name : string
        name of the zone
    nodataset : boolean
        do not create a ZFS file system
    brand_opts : boolean
        brand specific options to pass

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        if zones[name]["state"] == "configured":
            if __opts__["test"]:
                res_install = {"status": True}
            else:
                res_install = __salt__["zoneadm.install"](name, nodataset, brand_opts)
            ret["result"] = res_install["status"]
            if ret["result"]:
                ret["changes"][name] = "installed"
                ret["comment"] = "The zone {} was installed.".format(name)
            else:
                ret["comment"] = []
                ret["comment"].append("Failed to install zone {}!".format(name))
                if "message" in res_install:
                    ret["comment"].append(res_install["message"])
                ret["comment"] = "\n".join(ret["comment"])
        else:
            ret["result"] = True
            ret["comment"] = "zone {} already installed.".format(name)
    else:
        ret["result"] = False
        ret["comment"] = "zone {} is not configured!".format(name)

    return ret


def uninstalled(name):
    """
    Ensure zone is uninstalled

    name : string
        name of the zone

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    zones = __salt__["zoneadm.list"](installed=True, configured=True)
    if name in zones:
        if zones[name]["state"] != "configured":
            if __opts__["test"]:
                res_uninstall = {"status": True}
            else:
                res_uninstall = __salt__["zoneadm.uninstall"](name)
            ret["result"] = res_uninstall["status"]
            if ret["result"]:
                ret["changes"][name] = "uninstalled"
                ret["comment"] = "The zone {} was uninstalled.".format(name)
            else:
                ret["comment"] = []
                ret["comment"].append("Failed to uninstall zone {}!".format(name))
                if "message" in res_uninstall:
                    ret["comment"].append(res_uninstall["message"])
                ret["comment"] = "\n".join(ret["comment"])
        else:
            ret["result"] = True
            ret["comment"] = "zone {} already uninstalled.".format(name)
    else:
        ## note: a non existing zone is not installed, we do not consider this a failure
        ret["result"] = True
        ret["comment"] = "zone {} is not configured!".format(name)

    return ret


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



import logging

import salt.utils.minion
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class FakeThreadingClass:
    name = "thread-name"


class MinionUtilTestCase(TestCase):
    """
    TestCase for salt.utils.minion
    """

    def test__read_proc_file_multiprocessing_false(self):
        """
        test get_minion_pillar when
        target exists
        """

        opts = {"multiprocessing": False}
        proc_data = {
            "tgt_type": "glob",
            "jid": "20200310230030623022",
            "tgt": "minion",
            "pid": 12345,
            "ret": "",
            "user": "root",
            "arg": [10],
            "fun": "test.sleep",
        }

        fake_thread = FakeThreadingClass()
        fake_thread.name = "20200310230030623022-Job-20200310230030623022"

        with patch("os.getpid", MagicMock(return_value=12345)):
            with patch("salt.utils.files.fopen", mock_open(read_data=b"msgpack")):
                with patch("salt.payload.loads", MagicMock(return_value=proc_data)):
                    with patch(
                        "salt.utils.process.os_is_running", MagicMock(return_value=True)
                    ):
                        with patch(
                            "threading.enumerate", MagicMock(return_value=[fake_thread])
                        ):
                            with patch(
                                "salt.utils.minion._check_cmdline",
                                MagicMock(return_value=True),
                            ):
                                data = salt.utils.minion._read_proc_file(
                                    "/var/cache/salt/minion/proc/20200310230030623022",
                                    opts,
                                )
                                self.assertEqual(data, proc_data)

        opts = {"multiprocessing": False}
        proc_data = {
            "tgt_type": "glob",
            "jid": "20200310230030623022",
            "tgt": "minion",
            "pid": 12345,
            "ret": "",
            "user": "root",
            "arg": [10],
            "fun": "test.sleep",
        }

        fake_thread = FakeThreadingClass()
        fake_thread.name = "20200310230030623022"

        with patch("os.getpid", MagicMock(return_value=12345)):
            with patch("salt.utils.files.fopen", mock_open(read_data=b"msgpack")):
                with patch("salt.payload.loads", MagicMock(return_value=proc_data)):
                    with patch(
                        "salt.utils.process.os_is_running", MagicMock(return_value=True)
                    ):
                        with patch(
                            "threading.enumerate", MagicMock(return_value=[fake_thread])
                        ):
                            with patch(
                                "salt.utils.minion._check_cmdline",
                                MagicMock(return_value=True),
                            ):
                                data = salt.utils.minion._read_proc_file(
                                    "/var/cache/salt/minion/proc/20200310230030623022",
                                    opts,
                                )
                                self.assertEqual(data, proc_data)

        opts = {"multiprocessing": False}
        proc_data = {
            "tgt_type": "glob",
            "jid": "20200310230030623022",
            "tgt": "minion",
            "pid": 12345,
            "ret": "",
            "user": "root",
            "arg": [10],
            "fun": "test.sleep",
        }

        fake_thread = FakeThreadingClass()
        fake_thread.name = "20200310230030623022"

        with patch("os.getpid", MagicMock(return_value=12345)):
            with patch("salt.utils.files.fopen", mock_open(read_data=b"msgpack")):
                with patch("salt.payload.loads", MagicMock(return_value=proc_data)):
                    with patch(
                        "salt.utils.process.os_is_running", MagicMock(return_value=True)
                    ):
                        with patch(
                            "threading.enumerate", MagicMock(return_value=[fake_thread])
                        ):
                            with patch(
                                "salt.utils.minion._check_cmdline",
                                MagicMock(return_value=False),
                            ):
                                with patch("os.remove", MagicMock(return_value=True)):
                                    data = salt.utils.minion._read_proc_file(
                                        "/var/cache/salt/minion/proc/20200310230030623022",
                                        opts,
                                    )
                                    self.assertEqual(data, None)

    def test__read_proc_file_multiprocessing_true(self):
        """
        test get_minion_pillar when
        target exists
        """

        opts = {"multiprocessing": True}
        proc_data = {
            "tgt_type": "glob",
            "jid": "20200310230030623022",
            "tgt": "minion",
            "pid": 12345,
            "ret": "",
            "user": "root",
            "arg": [10],
            "fun": "test.sleep",
        }

        with patch("os.getpid", MagicMock(return_value=12345)):
            with patch("salt.utils.files.fopen", mock_open(read_data=b"msgpack")):
                with patch("salt.payload.loads", MagicMock(return_value=proc_data)):
                    with patch(
                        "salt.utils.process.os_is_running", MagicMock(return_value=True)
                    ):
                        with patch(
                            "salt.utils.minion._check_cmdline",
                            MagicMock(return_value=True),
                        ):
                            data = salt.utils.minion._read_proc_file(
                                "/var/cache/salt/minion/proc/20200310230030623022", opts
                            )
                            self.assertEqual(data, None)



import pytest
import salt.states.mac_xattr as xattr
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {xattr: {}}


def test_exists_not():
    """
    Test adding an attribute when it doesn't exist
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {"key": "value"},
            "comment": "",
            "name": "/path/to/file",
            "result": True,
        }

        exists_mock.return_value = True

        list_mock = MagicMock(return_value={"other.id": "value2"})
        write_mock = MagicMock()
        with patch.dict(
            xattr.__salt__, {"xattr.list": list_mock, "xattr.write": write_mock}
        ):
            out = xattr.exists("/path/to/file", ["key=value"])

            list_mock.assert_called_once_with("/path/to/file")
            write_mock.assert_called_once_with("/path/to/file", "key", "value", False)
            assert out == expected


def test_exists_change():
    """
    Test changing an attribute value
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {"key": "other_value"},
            "comment": "",
            "name": "/path/to/file",
            "result": True,
        }

        exists_mock.return_value = True

        list_mock = MagicMock(return_value={"key": "value"})
        write_mock = MagicMock()
        with patch.dict(
            xattr.__salt__, {"xattr.list": list_mock, "xattr.write": write_mock}
        ):
            out = xattr.exists("/path/to/file", ["key=other_value"])

            list_mock.assert_called_once_with("/path/to/file")
            write_mock.assert_called_once_with(
                "/path/to/file", "key", "other_value", False
            )
            assert out == expected


def test_exists_already():
    """
    Test with the same value does nothing
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {},
            "comment": "All values existed correctly.",
            "name": "/path/to/file",
            "result": True,
        }

        exists_mock.return_value = True

        list_mock = MagicMock(return_value={"key": "value"})
        write_mock = MagicMock()
        with patch.dict(
            xattr.__salt__, {"xattr.list": list_mock, "xattr.write": write_mock}
        ):
            out = xattr.exists("/path/to/file", ["key=value"])

            list_mock.assert_called_once_with("/path/to/file")
            assert not write_mock.called
            assert out == expected


def test_delete():
    """
    Test deleting an attribute from a file
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {"key": "delete"},
            "comment": "",
            "name": "/path/to/file",
            "result": True,
        }

        exists_mock.return_value = True

        list_mock = MagicMock(return_value={"key": "value2"})
        delete_mock = MagicMock()
        with patch.dict(
            xattr.__salt__, {"xattr.list": list_mock, "xattr.delete": delete_mock}
        ):
            out = xattr.delete("/path/to/file", ["key"])

            list_mock.assert_called_once_with("/path/to/file")
            delete_mock.assert_called_once_with("/path/to/file", "key")
            assert out == expected


def test_delete_not():
    """
    Test deleting an attribute that doesn't exist from a file
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {},
            "comment": "All attributes were already deleted.",
            "name": "/path/to/file",
            "result": True,
        }

        exists_mock.return_value = True

        list_mock = MagicMock(return_value={"other.key": "value2"})
        delete_mock = MagicMock()
        with patch.dict(
            xattr.__salt__, {"xattr.list": list_mock, "xattr.delete": delete_mock}
        ):
            out = xattr.delete("/path/to/file", ["key"])

            list_mock.assert_called_once_with("/path/to/file")
            assert not delete_mock.called
            assert out == expected



import pytest
import salt.client.ssh.client
import salt.utils.msgpack
from salt.client import ssh
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skip_if_binaries_missing("ssh", "ssh-keygen", check_all=True),
]


@pytest.fixture
def ssh_target(tmpdir):
    argv = [
        "ssh.set_auth_key",
        "root",
        "hobn+amNAXSBTiOXEqlBjGB...rsa root@master",
    ]

    opts = {
        "argv": argv,
        "__role": "master",
        "cachedir": tmpdir.strpath,
        "extension_modules": tmpdir.join("extmods").strpath,
    }
    target = {
        "passwd": "abc123",
        "ssh_options": None,
        "sudo": False,
        "identities_only": False,
        "host": "login1",
        "user": "root",
        "timeout": 65,
        "remote_port_forwards": None,
        "sudo_user": "",
        "port": "22",
        "priv": "/etc/salt/pki/master/ssh/salt-ssh.rsa",
    }
    return opts, target


@pytest.mark.skip_on_windows(reason="SSH_PY_SHIM not set on windows")
def test_cmd_block_python_version_error(ssh_target):
    opts = ssh_target[0]
    target = ssh_target[1]

    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        **target
    )
    mock_shim = MagicMock(
        return_value=(("", "ERROR: Unable to locate appropriate python command\n", 10))
    )
    patch_shim = patch("salt.client.ssh.Single.shim_cmd", mock_shim)
    with patch_shim:
        ret = single.cmd_block()
        assert "ERROR: Python version error. Recommendation(s) follow:" in ret[0]


@pytest.mark.parametrize(
    "test_opts",
    [
        ("extra_filerefs", "salt://foobar", True),
        ("host", "testhost", False),
        ("ssh_user", "testuser", True),
        ("ssh_passwd", "testpasswd", True),
        ("ssh_port", 23, False),
        ("ssh_sudo", True, True),
        ("ssh_sudo_user", "sudouser", False),
        ("ssh_priv", "test_priv", True),
        ("ssh_priv_passwd", "sshpasswd", True),
        ("ssh_identities_only", True, True),
        ("ssh_remote_port_forwards", "test", True),
        ("ssh_options", ["test1", "test2"], True),
        ("ssh_max_procs", 2, True),
        ("ssh_askpass", True, True),
        ("ssh_key_deploy", True, True),
        ("ssh_update_roster", True, True),
        ("ssh_scan_ports", "test", True),
        ("ssh_scan_timeout", 1.0, True),
        ("ssh_timeout", 1, False),
        ("ssh_log_file", "/tmp/test", True),
        ("raw_shell", True, True),
        ("refresh_cache", True, True),
        ("roster", "/test", True),
        ("roster_file", "/test1", True),
        ("rosters", ["test1"], False),
        ("ignore_host_keys", True, True),
        ("min_extra_mods", "test", True),
        ("thin_extra_mods", "test1", True),
        ("verbose", True, True),
        ("static", True, True),
        ("ssh_wipe", True, True),
        ("rand_thin_dir", True, True),
        ("regen_thin", True, True),
        ("ssh_run_pre_flight", True, True),
        ("no_host_keys", True, True),
        ("saltfile", "/tmp/test", True),
        ("doesnotexist", None, False),
    ],
)
def test_ssh_kwargs(test_opts):
    """
    test all ssh kwargs are not excluded from kwargs
    when preparing the SSH opts
    """
    opt_key = test_opts[0]
    opt_value = test_opts[1]
    # Is the kwarg in salt.utils.parsers?
    in_parser = test_opts[2]

    opts = {
        "eauth": "auto",
        "username": "test",
        "password": "test",
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        opt_key: opt_value,
    }
    client = salt.client.ssh.client.SSHClient(disable_custom_roster=True)
    if in_parser:
        ssh_kwargs = salt.utils.parsers.SaltSSHOptionParser().defaults
        assert opt_key in ssh_kwargs

    with patch("salt.roster.get_roster_file", MagicMock(return_value="")):
        ssh_obj = client._prep_ssh(**opts)
        assert ssh_obj.opts.get(opt_key, None) == opt_value



"""
Tests for salt.modules.linux_sysctl module

:codeauthor: jmoney <justin@saltstack.com>
"""

import pytest
import salt.modules.linux_sysctl as linux_sysctl
import salt.modules.systemd_service as systemd
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {linux_sysctl: {}, systemd: {}}


def test_get():
    """
    Tests the return of get function
    """
    mock_cmd = MagicMock(return_value=1)
    with patch.dict(linux_sysctl.__salt__, {"cmd.run": mock_cmd}):
        assert linux_sysctl.get("net.ipv4.ip_forward") == 1


def test_assign_proc_sys_failed():
    """
    Tests if /proc/sys/<kernel-subsystem> exists or not
    """
    with patch("os.path.exists", MagicMock(return_value=False)):
        cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 1",
        }
        mock_cmd = MagicMock(return_value=cmd)
        with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
            with pytest.raises(CommandExecutionError):
                linux_sysctl.assign("net.ipv4.ip_forward", 1)


def test_assign_cmd_failed():
    """
    Tests if the assignment was successful or not
    """
    with patch("os.path.exists", MagicMock(return_value=True)):
        cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": 'sysctl: setting key "net.ipv4.ip_forward": Invalid argument',
            "stdout": "net.ipv4.ip_forward = backward",
        }
        mock_cmd = MagicMock(return_value=cmd)
        with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
            with pytest.raises(CommandExecutionError):
                linux_sysctl.assign("net.ipv4.ip_forward", "backward")


def test_assign_success():
    """
    Tests the return of successful assign function
    """
    with patch("os.path.exists", MagicMock(return_value=True)):
        cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 1",
        }
        ret = {"net.ipv4.ip_forward": "1"}
        mock_cmd = MagicMock(return_value=cmd)
        with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
            assert linux_sysctl.assign("net.ipv4.ip_forward", 1) == ret


def test_persist_no_conf_failure():
    """
    Tests adding of config file failure
    """
    asn_cmd = {
        "pid": 1337,
        "retcode": 0,
        "stderr": "sysctl: permission denied",
        "stdout": "",
    }
    mock_asn_cmd = MagicMock(return_value=asn_cmd)
    cmd = "sysctl -w net.ipv4.ip_forward=1"
    mock_cmd = MagicMock(return_value=cmd)
    with patch.dict(
        linux_sysctl.__salt__,
        {"cmd.run_stdout": mock_cmd, "cmd.run_all": mock_asn_cmd},
    ):
        with patch("salt.utils.files.fopen", mock_open()) as m_open:
            with pytest.raises(CommandExecutionError):
                linux_sysctl.persist("net.ipv4.ip_forward", 1, config=None)


def test_persist_no_conf_success():
    """
    Tests successful add of config file when previously not one
    """
    config = "/etc/sysctl.conf"
    with patch("os.path.isfile", MagicMock(return_value=False)), patch(
        "os.path.exists", MagicMock(return_value=True)
    ):
        asn_cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 1",
        }
        mock_asn_cmd = MagicMock(return_value=asn_cmd)

        sys_cmd = "systemd 208\n+PAM +LIBWRAP"
        mock_sys_cmd = MagicMock(return_value=sys_cmd)

        with patch("salt.utils.files.fopen", mock_open()) as m_open, patch.dict(
            linux_sysctl.__context__, {"salt.utils.systemd.version": 232}
        ), patch.dict(
            linux_sysctl.__salt__,
            {"cmd.run_stdout": mock_sys_cmd, "cmd.run_all": mock_asn_cmd},
        ), patch.dict(
            systemd.__context__,
            {"salt.utils.systemd.booted": True, "salt.utils.systemd.version": 232},
        ):
            linux_sysctl.persist("net.ipv4.ip_forward", 1, config=config)
            writes = m_open.write_calls()
            assert writes == ["#\n# Kernel sysctl configuration\n#\n"], writes


def test_persist_read_conf_success():
    """
    Tests sysctl.conf read success
    """
    with patch("os.path.isfile", MagicMock(return_value=True)), patch(
        "os.path.exists", MagicMock(return_value=True)
    ):
        asn_cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 1",
        }
        mock_asn_cmd = MagicMock(return_value=asn_cmd)

        sys_cmd = "systemd 208\n+PAM +LIBWRAP"
        mock_sys_cmd = MagicMock(return_value=sys_cmd)

        with patch("salt.utils.files.fopen", mock_open()):
            with patch.dict(
                linux_sysctl.__context__, {"salt.utils.systemd.version": 232}
            ):
                with patch.dict(
                    linux_sysctl.__salt__,
                    {"cmd.run_stdout": mock_sys_cmd, "cmd.run_all": mock_asn_cmd},
                ):
                    with patch.dict(
                        systemd.__context__, {"salt.utils.systemd.booted": True}
                    ):
                        assert (
                            linux_sysctl.persist("net.ipv4.ip_forward", 1) == "Updated"
                        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.keyboard as keyboard
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class KeyboardTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.keyboard
    """

    def setup_loader_modules(self):
        return {keyboard: {}}

    # 'get_sys' function tests: 1

    def test_get_sys(self):
        """
        Test if it get current system keyboard setting
        """
        mock = MagicMock(return_value="X11 Layout=us")
        with patch.dict(keyboard.__grains__, {"os_family": "RedHat"}):
            with patch.dict(keyboard.__salt__, {"cmd.run": mock}):
                self.assertEqual(keyboard.get_sys(), "us")

    # 'set_sys' function tests: 1

    def test_set_sys(self):
        """
        Test if it set current system keyboard setting
        """
        mock = MagicMock(return_value="us")
        with patch.dict(keyboard.__grains__, {"os_family": "RedHat"}):
            with patch.dict(keyboard.__salt__, {"cmd.run": mock}):
                with patch.dict(keyboard.__salt__, {"file.sed": MagicMock()}):
                    self.assertEqual(keyboard.set_sys("us"), "us")

    # 'get_x' function tests: 1

    def test_get_x(self):
        """
        Test if it get current X keyboard setting
        """
        mock = MagicMock(return_value="layout:     us")
        with patch.dict(keyboard.__salt__, {"cmd.run": mock}):
            self.assertEqual(keyboard.get_x(), "us")

    # 'set_x' function tests: 1

    def test_set_x(self):
        """
        Test if it set current X keyboard setting
        """
        mock = MagicMock(return_value="us")
        with patch.dict(keyboard.__salt__, {"cmd.run": mock}):
            self.assertEqual(keyboard.set_x("us"), "us")



"""
Support for IFTTT

.. versionadded:: 2015.8.0

Requires an ``api_key`` in ``/etc/salt/minion``:

.. code-block:: yaml

    ifttt:
      secret_key: '280d4699-a817-4719-ba6f-ca56e573e44f'
"""


import logging
import time

import salt.utils.http
import salt.utils.json

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load the module if apache is installed
    """
    if not __salt__["config.get"]("ifttt.secret_key") and not __salt__["config.get"](
        "ifttt:secret_key"
    ):
        return (False, "IFTTT Secret Key Unavailable, not loading.")
    return True


def _query(event=None, method="GET", args=None, header_dict=None, data=None):
    """
    Make a web call to IFTTT.
    """
    secret_key = __salt__["config.get"]("ifttt.secret_key") or __salt__["config.get"](
        "ifttt:secret_key"
    )
    path = "https://maker.ifttt.com/trigger/{}/with/key/{}".format(event, secret_key)

    if header_dict is None:
        header_dict = {"Content-type": "application/json"}

    if method != "POST":
        header_dict["Accept"] = "application/json"

    result = salt.utils.http.query(
        path,
        method,
        params={},
        data=data,
        header_dict=header_dict,
        decode=True,
        decode_type="auto",
        text=True,
        status=True,
        cookies=True,
        persist_session=True,
        opts=__opts__,
        backend="requests",
    )
    return result


def trigger_event(event=None, **kwargs):
    """
    Trigger a configured event in IFTTT.

    :param event:   The name of the event to trigger.

    :return:        A dictionary with status, text, and error if result was failure.
    """

    res = {"result": False, "message": "Something went wrong"}

    data = {}
    for value in ("value1", "value2", "value3", "Value1", "Value2", "Value3"):
        if value in kwargs:
            data[value.lower()] = kwargs[value]
    data["occurredat"] = time.strftime("%B %d, %Y %I:%M%p", time.localtime())
    result = _query(event=event, method="POST", data=salt.utils.json.dumps(data))
    if "status" in result:
        if result["status"] == 200:
            res["result"] = True
            res["message"] = result["text"]
        else:
            if "error" in result:
                res["message"] = result["error"]
    return res



"""
Compendium of generic DNS utilities.
The 'dig' command line tool must be installed in order to use this module.
"""

import logging
import re

import salt.utils.network
import salt.utils.path

log = logging.getLogger(__name__)

__virtualname__ = "dig"


def __virtual__():
    """
    Only load module if dig binary is present
    """
    if salt.utils.path.which("dig"):
        return __virtualname__
    return (
        False,
        "The dig execution module cannot be loaded: the dig binary is not in the path.",
    )


def check_ip(addr):
    """
    Check if address is a valid IP. returns True if valid, otherwise False.

    CLI Example:

    .. code-block:: bash

        salt ns1 dig.check_ip 127.0.0.1
        salt ns1 dig.check_ip 1111:2222:3333:4444:5555:6666:7777:8888
    """

    try:
        addr = addr.rsplit("/", 1)
    except AttributeError:
        # Non-string passed
        return False

    if salt.utils.network.is_ipv4(addr[0]):
        try:
            if 1 <= int(addr[1]) <= 32:
                return True
        except ValueError:
            # Non-int subnet notation
            return False
        except IndexError:
            # No subnet notation used (i.e. just an IPv4 address)
            return True

    if salt.utils.network.is_ipv6(addr[0]):
        try:
            if 8 <= int(addr[1]) <= 128:
                return True
        except ValueError:
            # Non-int subnet notation
            return False
        except IndexError:
            # No subnet notation used (i.e. just an IPv4 address)
            return True

    return False


def A(host, nameserver=None):
    """
    Return the A record for ``host``.

    Always returns a list.

    CLI Example:

    .. code-block:: bash

        salt ns1 dig.A www.google.com
    """
    dig = ["dig", "+short", str(host), "A"]

    if nameserver is not None:
        dig.append("@{}".format(nameserver))

    cmd = __salt__["cmd.run_all"](dig, python_shell=False)
    # In this case, 0 is not the same as False
    if cmd["retcode"] != 0:
        log.warning(
            "dig returned exit code '%s'. Returning empty list as fallback.",
            cmd["retcode"],
        )
        return []

    # make sure all entries are IPs
    return [x for x in cmd["stdout"].split("\n") if check_ip(x)]


def AAAA(host, nameserver=None):
    """
    Return the AAAA record for ``host``.

    Always returns a list.

    CLI Example:

    .. code-block:: bash

        salt ns1 dig.AAAA www.google.com
    """
    dig = ["dig", "+short", str(host), "AAAA"]

    if nameserver is not None:
        dig.append("@{}".format(nameserver))

    cmd = __salt__["cmd.run_all"](dig, python_shell=False)
    # In this case, 0 is not the same as False
    if cmd["retcode"] != 0:
        log.warning(
            "dig returned exit code '%s'. Returning empty list as fallback.",
            cmd["retcode"],
        )
        return []

    # make sure all entries are IPs
    return [x for x in cmd["stdout"].split("\n") if check_ip(x)]


def NS(domain, resolve=True, nameserver=None):
    """
    Return a list of IPs of the nameservers for ``domain``

    If ``resolve`` is False, don't resolve names.

    CLI Example:

    .. code-block:: bash

        salt ns1 dig.NS google.com
    """
    dig = ["dig", "+short", str(domain), "NS"]

    if nameserver is not None:
        dig.append("@{}".format(nameserver))

    cmd = __salt__["cmd.run_all"](dig, python_shell=False)
    # In this case, 0 is not the same as False
    if cmd["retcode"] != 0:
        log.warning(
            "dig returned exit code '%s'. Returning empty list as fallback.",
            cmd["retcode"],
        )
        return []

    if resolve:
        ret = []
        for ns_host in cmd["stdout"].split("\n"):
            for ip_addr in A(ns_host, nameserver):
                ret.append(ip_addr)
        return ret

    return cmd["stdout"].split("\n")


def SPF(domain, record="SPF", nameserver=None):
    """
    Return the allowed IPv4 ranges in the SPF record for ``domain``.

    If record is ``SPF`` and the SPF record is empty, the TXT record will be
    searched automatically. If you know the domain uses TXT and not SPF,
    specifying that will save a lookup.

    CLI Example:

    .. code-block:: bash

        salt ns1 dig.SPF google.com
    """
    spf_re = re.compile(r"(?:\+|~)?(ip[46]|include):(.+)")
    cmd = ["dig", "+short", str(domain), record]

    if nameserver is not None:
        cmd.append("@{}".format(nameserver))

    result = __salt__["cmd.run_all"](cmd, python_shell=False)
    # In this case, 0 is not the same as False
    if result["retcode"] != 0:
        log.warning(
            "dig returned exit code '%s'. Returning empty list as fallback.",
            result["retcode"],
        )
        return []

    if result["stdout"] == "" and record == "SPF":
        # empty string is successful query, but nothing to return. So, try TXT
        # record.
        return SPF(domain, "TXT", nameserver)

    sections = re.sub('"', "", result["stdout"]).split()
    if not sections or sections[0] != "v=spf1":
        return []

    if sections[1].startswith("redirect="):
        # Run a lookup on the part after 'redirect=' (9 chars)
        return SPF(sections[1][9:], "SPF", nameserver)
    ret = []
    for section in sections[1:]:
        try:
            mechanism, address = spf_re.match(section).groups()
        except AttributeError:
            # Regex was not matched
            continue
        if mechanism == "include":
            ret.extend(SPF(address, "SPF", nameserver))
        elif mechanism in ("ip4", "ip6") and check_ip(address):
            ret.append(address)
    return ret


def MX(domain, resolve=False, nameserver=None):
    """
    Return a list of lists for the MX of ``domain``.

    If the ``resolve`` argument is True, resolve IPs for the servers.

    It's limited to one IP, because although in practice it's very rarely a
    round robin, it is an acceptable configuration and pulling just one IP lets
    the data be similar to the non-resolved version. If you think an MX has
    multiple IPs, don't use the resolver here, resolve them in a separate step.

    CLI Example:

    .. code-block:: bash

        salt ns1 dig.MX google.com
    """
    dig = ["dig", "+short", str(domain), "MX"]

    if nameserver is not None:
        dig.append("@{}".format(nameserver))

    cmd = __salt__["cmd.run_all"](dig, python_shell=False)
    # In this case, 0 is not the same as False
    if cmd["retcode"] != 0:
        log.warning(
            "dig returned exit code '%s'. Returning empty list as fallback.",
            cmd["retcode"],
        )
        return []

    stdout = [x.split() for x in cmd["stdout"].split("\n")]

    if resolve:
        return [(lambda x: [x[0], A(x[1], nameserver)[0]])(x) for x in stdout]

    return stdout


def TXT(host, nameserver=None):
    """
    Return the TXT record for ``host``.

    Always returns a list.

    CLI Example:

    .. code-block:: bash

        salt ns1 dig.TXT google.com
    """
    dig = ["dig", "+short", str(host), "TXT"]

    if nameserver is not None:
        dig.append("@{}".format(nameserver))

    cmd = __salt__["cmd.run_all"](dig, python_shell=False)

    if cmd["retcode"] != 0:
        log.warning(
            "dig returned exit code '%s'. Returning empty list as fallback.",
            cmd["retcode"],
        )
        return []

    return [i for i in cmd["stdout"].split("\n")]


# Let lowercase work, since that is the convention for Salt functions
a = A
aaaa = AAAA
ns = NS
spf = SPF
mx = MX



"""
Generate baseline proxy minion grains for ESXi hosts.

.. versionadded:: 2015.8.4

"""


import logging

import salt.utils.proxy
from salt.exceptions import SaltSystemExit

__proxyenabled__ = ["esxi"]
__virtualname__ = "esxi"

log = logging.getLogger(__file__)

GRAINS_CACHE = {}


def __virtual__():

    # import salt.utils.proxy again
    # so it is available for tests.
    import salt.utils.proxy

    try:
        if salt.utils.proxy.is_proxytype(__opts__, "esxi"):
            import salt.modules.vsphere

            return __virtualname__
    except KeyError:
        pass

    return False


def esxi():
    return _grains()


def kernel():
    return {"kernel": "proxy"}


def os():
    if not GRAINS_CACHE:
        GRAINS_CACHE.update(_grains())

    try:
        return {"os": GRAINS_CACHE.get("fullName")}
    except AttributeError:
        return {"os": "Unknown"}


def os_family():
    return {"os_family": "proxy"}


def _find_credentials(host):
    """
    Cycle through all the possible credentials and return the first one that
    works.
    """
    user_names = [__pillar__["proxy"].get("username", "root")]
    passwords = __pillar__["proxy"]["passwords"]
    for user in user_names:
        for password in passwords:
            try:
                # Try to authenticate with the given user/password combination
                ret = salt.modules.vsphere.system_info(
                    host=host, username=user, password=password
                )
            except SaltSystemExit:
                # If we can't authenticate, continue on to try the next password.
                continue
            # If we have data returned from above, we've successfully authenticated.
            if ret:
                return user, password
    # We've reached the end of the list without successfully authenticating.
    raise SaltSystemExit(
        "Cannot complete login due to an incorrect user name or password."
    )


def _grains():
    """
    Get the grains from the proxied device.
    """
    try:
        host = __pillar__["proxy"]["host"]
        if host:
            username, password = _find_credentials(host)
            protocol = __pillar__["proxy"].get("protocol")
            port = __pillar__["proxy"].get("port")
            ret = salt.modules.vsphere.system_info(
                host=host,
                username=username,
                password=password,
                protocol=protocol,
                port=port,
            )
            GRAINS_CACHE.update(ret)
    except KeyError:
        pass

    return GRAINS_CACHE
